import os
from typing import Tuple

import pytest

from .adapters.urdf import URDFParser
from .connections import Connection6DoF, PrismaticConnection, RevoluteConnection, FixedConnection, OmniDrive
from .geometry import Box, Scale, Sphere
from .prefixed_name import PrefixedName
from .spatial_types import TransformationMatrix
from .spatial_types.derivatives import Derivatives, DerivativeMap
from .spatial_types.spatial_types import Vector3
from .world import World
from .world_entity import Body

@pytest.fixture
def world_setup() -> Tuple[World, Body, Body, Body, Body, Body]:
    world = World()
    root = Body(name=PrefixedName(name='root', prefix='world'))
    l1 = Body(name=PrefixedName('l1'))
    l2 = Body(name=PrefixedName('l2'))
    bf = Body(name=PrefixedName('bf'))
    r1 = Body(name=PrefixedName('r1'))
    r2 = Body(name=PrefixedName('r2'))

    with world.modify_world():
        [world.add_body(b) for b in [root, l1, l2, bf, r1, r2]]
        lower_limits = DerivativeMap()
        lower_limits.velocity = -1
        upper_limits = DerivativeMap()
        upper_limits.velocity = 1
        dof = world.create_degree_of_freedom(name=PrefixedName('dof'), lower_limits=lower_limits,
                                             upper_limits=upper_limits)

        c_l1_l2 = PrismaticConnection(parent=l1, child=l2, dof=dof, axis=Vector3.X(reference_frame=l1))
        c_r1_r2 = RevoluteConnection(parent=r1, child=r2, dof=dof, axis=Vector3.Z(reference_frame=r1))
        bf_root_l1 = FixedConnection(parent=bf, child=l1)
        bf_root_r1 = FixedConnection(parent=bf, child=r1)
        world.add_connection(c_l1_l2)
        world.add_connection(c_r1_r2)
        world.add_connection(bf_root_l1)
        world.add_connection(bf_root_r1)
        c_root_bf = Connection6DoF(parent=root, child=bf, _world=world)
        world.add_connection(c_root_bf)

    return world, l1, l2, bf, r1, r2

@pytest.fixture
def world_setup_simple():
    world = World()
    root = Body(name=PrefixedName(name='root', prefix='world'))
    body1 = Body(name=PrefixedName('name1', prefix='test'),
                 collision=[Box(origin=TransformationMatrix.from_xyz_rpy(), scale=Scale(0.25, 0.25, 0.25))])
    body2 = Body(name=PrefixedName('name2', prefix='test'),
                 collision=[Box(origin=TransformationMatrix.from_xyz_rpy(), scale=Scale(0.25, 0.25, 0.25))])
    body3 = Body(name=PrefixedName('name3', prefix='test'),
                 collision=[Sphere(origin=TransformationMatrix.from_xyz_rpy(), radius=0.01)])
    body4 = Body(name=PrefixedName('name4', prefix='test'),
                 collision=[Sphere(origin=TransformationMatrix.from_xyz_rpy(), radius=0.01)])

    with world.modify_world():
        world.add_body(body1)
        world.add_body(body2)
        world.add_body(body3)
        world.add_body(body4)

        c_root_body1 = Connection6DoF(parent=root, child=body1, _world=world)
        c_root_body2 = Connection6DoF(parent=root, child=body2, _world=world)
        c_root_body3 = Connection6DoF(parent=root, child=body3, _world=world)
        c_root_body4 = Connection6DoF(parent=root, child=body4, _world=world)

        world.add_connection(c_root_body1)
        world.add_connection(c_root_body2)
        world.add_connection(c_root_body3)
        world.add_connection(c_root_body4)
    return world, body1, body2, body3, body4


@pytest.fixture
def two_arm_robot_world():
    urdf_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf")
    robot = os.path.join(urdf_dir, "simple_two_arm_robot.urdf")
    world = World()
    with world.modify_world():
        localization_body = Body(name=PrefixedName('odom_combined'))
        world.add_body(localization_body)

        robot_parser = URDFParser(robot)
        world_with_robot = robot_parser.parse()
        # world_with_pr2.plot_kinematic_structure()
        root = world_with_robot.root
        c_root_bf = OmniDrive(parent=localization_body, child=root, _world=world)
        world.merge_world(world_with_robot, root_connection=c_root_bf)
    return world

@pytest.fixture
def pr2_world():
    urdf_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf")
    pr2 = os.path.join(urdf_dir, "pr2_kinematic_tree.urdf")
    world = World()
    with world.modify_world():
        localization_body = Body(name=PrefixedName('odom_combined'))
        world.add_body(localization_body)

        pr2_parser = URDFParser(pr2)
        world_with_pr2 = pr2_parser.parse()
        # world_with_pr2.plot_kinematic_structure()
        pr2_root = world_with_pr2.root
        c_root_bf = OmniDrive(parent=localization_body, child=pr2_root, _world=world)
        world.merge_world(world_with_pr2, c_root_bf)

    return world




import os
from dataclasses import dataclass
from typing import Optional, Tuple, Dict, Union, List
from ..spatial_types import spatial_types as cas
from urdf_parser_py import urdf

from ..connections import RevoluteConnection, PrismaticConnection, FixedConnection
from ..prefixed_name import PrefixedName
from ..spatial_types.derivatives import Derivatives, DerivativeMap
from ..spatial_types.spatial_types import TransformationMatrix, Vector3
from ..utils import suppress_stdout_stderr, hacky_urdf_parser_fix
from ..world import World, Body, Connection
from ..geometry import Box, Sphere, Cylinder, Mesh, Scale, Shape, Color

connection_type_map = {  # 'unknown': JointType.UNKNOWN,
    'revolute': RevoluteConnection,
    'continuous': RevoluteConnection,
    'prismatic': PrismaticConnection,
    # 'floating': JointType.FLOATING,
    # 'planar': JointType.PLANAR,
    'fixed': FixedConnection}


def urdf_joint_to_limits(urdf_joint: urdf.Joint) -> Tuple[DerivativeMap[float], DerivativeMap[float]]:
    lower_limits = DerivativeMap()
    upper_limits = DerivativeMap()
    if not urdf_joint.type == 'continuous':
        try:
            lower_limits.position = max(urdf_joint.safety_controller.soft_lower_limit,
                                                     urdf_joint.limit.lower)
            upper_limits.position = min(urdf_joint.safety_controller.soft_upper_limit,
                                                     urdf_joint.limit.upper)
        except AttributeError:
            try:
                lower_limits.position = urdf_joint.limit.lower
                upper_limits.position = urdf_joint.limit.upper
            except AttributeError:
                pass
    try:
        lower_limits.velocity = -urdf_joint.limit.velocity
        upper_limits.velocity = urdf_joint.limit.velocity
    except AttributeError:
        pass
    if urdf_joint.mimic is not None:
        if urdf_joint.mimic.multiplier is not None:
            multiplier = urdf_joint.mimic.multiplier
        else:
            multiplier = 1
        if urdf_joint.mimic.offset is not None:
            offset = urdf_joint.mimic.offset
        else:
            offset = 0
        for d2 in Derivatives.range(Derivatives.position, Derivatives.velocity):
            lower_limits.data[d2] -= offset
            upper_limits.data[d2] -= offset
            if multiplier < 0:
                upper_limits.data[d2], lower_limits.data[d2] = lower_limits.data[d2], upper_limits.data[d2]
            upper_limits.data[d2] /= multiplier
            lower_limits.data[d2] /= multiplier
    return lower_limits, upper_limits


@dataclass
class URDFParser:
    """
    Class to parse URDF files to worlds.
    """

    file_path: str
    """
    The file path of the URDF.
    """

    prefix: Optional[str] = None
    """
    The prefix for every name used in this world.
    """

    def __post_init__(self):
        if self.prefix is None:
            self.prefix = os.path.basename(self.file_path).split('.')[0]

    def parse(self) -> World:
        # cache_dir = os.path.join(os.getcwd(), '..', '..', '../resources', 'cache')
        # file_name = os.path.basename(self.file_path)
        # new_file_path = os.path.join(cache_dir, file_name)
        # generate_from_description_file(self.file_path, new_file_path)

        with open(self.file_path, 'r') as file:
            # Since parsing URDF causes a lot of warning messages which can't be deactivated, we suppress them
            with suppress_stdout_stderr():
                self.parsed = urdf.URDF.from_xml_string(hacky_urdf_parser_fix(file.read()))

        links = [self.parse_link(link, PrefixedName(link.name, self.parsed.name)) for link in self.parsed.links]
        root = [link for link in links if link.name.name == self.parsed.get_root()][0]
        world = World()
        world.name = self.prefix
        world.add_body(root)

        with world.modify_world():
            joints = []
            for joint in self.parsed.joints:
                parent = [link for link in links if link.name.name == joint.parent][0]
                child = [link for link in links if link.name.name == joint.child][0]
                parsed_joint = self.parse_joint(joint, parent, child, world)
                joints.append(parsed_joint)

            [world.add_connection(joint) for joint in joints]
            [world.add_body(link) for link in links]

        return world

    def parse_joint(self, joint: urdf.Joint, parent: Body, child: Body, world: World) -> Connection:
        connection_type = connection_type_map.get(joint.type, Connection)
        if joint.origin is not None:
            translation_offset = joint.origin.xyz
            rotation_offset = joint.origin.rpy
        else:
            translation_offset = None
            rotation_offset = None
        if translation_offset is None:
            translation_offset = [0, 0, 0]
        if rotation_offset is None:
            rotation_offset = [0, 0, 0]
        parent_T_child = cas.TransformationMatrix.from_xyz_rpy(x=translation_offset[0],
                                                               y=translation_offset[1],
                                                               z=translation_offset[2],
                                                               roll=rotation_offset[0],
                                                               pitch=rotation_offset[1],
                                                               yaw=rotation_offset[2])
        if connection_type == FixedConnection:
            return connection_type(parent=parent, child=child, origin_expression=parent_T_child)

        lower_limits, upper_limits = urdf_joint_to_limits(joint)
        is_mimic = joint.mimic is not None
        multiplier = None
        offset = None
        if is_mimic:
            if joint.mimic.multiplier is not None:
                multiplier = joint.mimic.multiplier
            else:
                multiplier = 1
            if joint.mimic.offset is not None:
                offset = joint.mimic.offset
            else:
                offset = 0

            dof_name = PrefixedName(joint.mimic.joint)
        else:
            dof_name = PrefixedName(joint.name)

        try:
            dof = world.get_degree_of_freedom_by_name(dof_name)
        except KeyError as e:
            dof = world.create_degree_of_freedom(name=PrefixedName(joint.name),
                                                 lower_limits=lower_limits, upper_limits=upper_limits)

        result = connection_type(parent=parent, child=child, origin_expression=parent_T_child,
                                 multiplier=multiplier, offset=offset,
                                 axis=Vector3(*map(int, joint.axis),
                                                 reference_frame=parent),
                                 dof=dof)
        return result

    def parse_link(self, link: urdf.Link, parent_frame: PrefixedName) -> Body:
        """
        Parses a URDF link to a link object.
        :param link: The URDF link to parse.
        :param parent_frame: The parent frame of the link, used for transformations of collisions and visuals.
        :return: The parsed link object.
        """
        name = PrefixedName(prefix=self.prefix, name=link.name)
        visuals = self.parse_geometry(link.visuals, parent_frame)
        collisions = self.parse_geometry(link.collisions, parent_frame)
        return Body(name=name, visual=visuals, collision=collisions)

    def parse_geometry(self, geometry: Union[List[urdf.Collision], List[urdf.Visual]], parent_frame: PrefixedName) -> \
            List[Shape]:
        """
        Parses a URDF geometry to the corresponding shapes.
        :param geometry: The URDF geometry to parse either the collisions of visuals.'
        :param parent_frame: The parent frame of the geometry, used for transformations.
        :return: A List of shapes corresponding to the URDF geometry.
        """
        res = []
        material_dict = dict(zip([material.name for material in self.parsed.materials],
                                 [material.color.rgba if material.color else None for material in
                                  self.parsed.materials]))
        for i, geom in enumerate(geometry):
            params = (*(geom.origin.xyz + geom.origin.rpy),) if geom.origin else (0, 0, 0, 0, 0, 0,)
            origin_transform = TransformationMatrix.from_xyz_rpy(*params)
            if isinstance(geom.geometry, urdf.Box):
                color = Color(*material_dict.get(geom.material.name,
                                                 (1, 1, 1, 1))) if hasattr(geom,
                                                                           "material") and geom.material else Color(
                    1,
                    1,
                    1,
                    1)
                res.append(Box(origin=origin_transform, scale=Scale(*geom.geometry.size), color=color))
            elif isinstance(geom.geometry, urdf.Sphere):
                color = Color(*material_dict.get(geom.material.name,
                                                 (1, 1, 1, 1))) if hasattr(geom,
                                                                           "material") and geom.material else Color(
                    1,
                    1,
                    1,
                    1)
                res.append(Sphere(origin=origin_transform, radius=geom.geometry.radius, color=color))
            elif isinstance(geom.geometry, urdf.Cylinder):
                color = Color(*material_dict.get(geom.material.name,
                                                 (1, 1, 1, 1))) if hasattr(geom,
                                                                           "material") and geom.material else Color(
                    1,
                    1,
                    1,
                    1)
                res.append(Cylinder(origin=origin_transform, width=geom.geometry.radius, height=geom.geometry.length,
                                    color=color))
            elif isinstance(geom.geometry, urdf.Mesh):
                if geom.geometry.filename is None:
                    raise ValueError("Mesh geometry must have a filename.")
                res.append(Mesh(origin=origin_transform, filename=self.parse_file_path(geom.geometry.filename),
                                scale=Scale(*(geom.geometry.scale or (1, 1, 1)))))
        return res

    @staticmethod
    def parse_file_path(file_path: str) -> str:
        """
        Parses a file path which contains a ros package to a path in the local file system.
        :param file_path: The path to the URDF file.
        :return: The parsed and processed file path.
        """
        if "package://" in file_path:
            try:
                from ament_index_python.packages import get_package_share_directory
            except ImportError:
                raise ImportError("No ROS install found while the URDF file contains references to ROS packages. ")
            # Splits the file path at '//' to get the package  and the rest of the path
            package_split = file_path.split('//')
            # Splits the path after the // to get the package name and the rest of the path
            package_name = package_split[1].split('/')[0]
            package_path = get_package_share_directory(package_name)
            file_path = file_path.replace("package://" + package_name, package_path)
        if 'file://' in file_path:
            file_path = file_path.replace("file://", './')
        return file_path



from __future__ import annotations

import logging

import matplotlib.pyplot as plt

from .geometry import BoundingBox, BoundingBoxCollection
from .variables import SpatialVariables
from .world import World
from .world_entity import View, EnvironmentView

logger = logging.getLogger(__name__)

import time
from functools import reduce
from operator import or_
from typing import List, Optional, Dict
# typing.Self is available starting with Python 3.11
from typing_extensions import Self

import numpy as np
import plotly.graph_objects as go
import rustworkx as rx
from random_events.interval import reals
from random_events.product_algebra import SimpleEvent, Event
from rtree import index
from sortedcontainers import SortedSet

from .spatial_types import Point3


class PoseOccupiedError(Exception):
    """
    Error that is raised when a pose is occupied or not in the search space of a Connectivity Graphs.
    """

    def __init__(self, point: Point3):
        """
        Construct a new pose occupied error.
        :param pose: The pose that is occupied.
        """
        super().__init__(f"The pose {point} is occupied.")
        self.point = point


class GraphOfConvexSets:
    """
    A graph that represents the connectivity between convex sets.

    Every node in the graph is a convex set, represented by a bounding box.
    Every edge in the graph represents the connectivity between two convex sets.
    """

    search_space: BoundingBoxCollection
    """
    The bounding box of the search space. Defaults to the entire three dimensional space.
    """

    graph: rx.PyGraph[BoundingBox]
    """
    The connectivity graph of the convex sets.
    """

    box_to_index_map: Dict[BoundingBox, int]
    """
    A mapping from bounding boxes to their indices in the graph.
    """

    def __init__(self, search_space: Optional[BoundingBoxCollection] = None):
        self.search_space = self._make_search_space(search_space)
        self.graph = rx.PyGraph(multigraph=False)
        self.box_to_index_map = {}

    def add_node(self, box: BoundingBox):
        self.box_to_index_map[box] = self.graph.add_node(box)

    def calculate_connectivity(self, tolerance=0.001):
        """
        Calculate the connectivity of the graph by checking for intersections between the bounding boxes of the nodes.
        This uses an R-tree for efficient spatial indexing and intersection queries.

        :param tolerance: The tolerance for the intersection when calculating the connectivity.
        """

        def _overlap(a_min, a_max, b_min, b_max) -> bool:
            return (a_min[0] <= b_max[0] and b_min[0] <= a_max[0] and
                    a_min[1] <= b_max[1] and b_min[1] <= a_max[1] and
                    a_min[2] <= b_max[2] and b_min[2] <= a_max[2])

        def _intersection_box(a_min, a_max, b_min, b_max):
            return BoundingBox(
                max(a_min[0], b_min[0]), max(a_min[1], b_min[1]), max(a_min[2], b_min[2]),
                min(a_max[0], b_max[0]), min(a_max[1], b_max[1]), min(a_max[2], b_max[2]),
            )

        # Build a 3-D R-tree
        prop = index.Property()
        prop.dimension = 3
        rtree_idx = index.Index(properties=prop)

        node_list = list(self.graph.nodes())
        orig_mins, orig_maxs, expanded = [], [], []

        # Record every node once, insert it into the index
        for n in node_list:
            mn = (n.min_x, n.min_y, n.min_z)
            mx = (n.max_x, n.max_y, n.max_z)
            ex = (mn[0] - tolerance, mn[1] - tolerance, mn[2] - tolerance,
                  mx[0] + tolerance, mx[1] + tolerance, mx[2] + tolerance)

            orig_mins.append(mn)
            orig_maxs.append(mx)
            expanded.append(ex)
            rtree_idx.insert(len(orig_mins) - 1, ex)

        # Query & link, skip self-loops and symmetric pairs
        for i, (mn_i, mx_i, ex_i) in enumerate(zip(orig_mins, orig_maxs, expanded)):
            for j in rtree_idx.intersection(ex_i):
                if j <= i:  # symmetry → skip
                    continue
                mn_j, mx_j = orig_mins[j], orig_maxs[j]
                if not _overlap(mn_i, mx_i, mn_j, mx_j):
                    continue  # no true overlap
                box = _intersection_box(mn_i, mx_i, mn_j, mx_j)

                self.graph.add_edge(i, j, box)

    def draw(self):
        import rustworkx.visualization
        rustworkx.visualization.mpl_draw(self.graph)
        plt.show()

    def plot_free_space(self) -> List[go.Mesh3d]:
        """
        Plot the free space of the environment in blue.
        :return: A list of traces that can be put into a plotly figure.
        """
        free_space = Event(*[node.simple_event for node in self.graph.nodes()])
        return free_space.plot(color="blue")

    def plot_occupied_space(self) -> List[go.Mesh3d]:
        """
        Plot the occupied space of the environment in red.
        :return: A list of traces that can be put into a plotly figure.
        """
        free_space = Event(*[node.simple_event for node in self.graph.nodes()])
        occupied_space = ~free_space & self.search_space.event
        return occupied_space.plot(color="red")

    def node_of_point(self, point: Point3) -> Optional[BoundingBox]:
        """
        Find the node that contains a point.

        :return: The node that contains the point or None if no node contains the point.
        """
        for node in self.graph.nodes():
            if node.contains(point):
                return node
        return None

    def path_from_to(self, start: Point3, goal: Point3) -> Optional[List[Point3]]:
        """
        Calculate a connected path from a start pose to a goal pose.

        :param start: The start pose.
        :param goal: The goal pose.
        :return: The path as a sequence of points to navigate to or None if no path exists.
        """

        # get poses from params
        start_node = self.node_of_point(start)
        goal_node = self.node_of_point(goal)

        # validate if the poses are part of the graph
        if start_node is None:
            raise PoseOccupiedError(start)
        if goal_node is None:
            raise PoseOccupiedError(goal)

        if start_node == goal_node:
            return [start, goal]

        # get the shortest path (perhaps replace with a*?)
        paths = rx.all_shortest_paths(self.graph, self.box_to_index_map[start_node], self.box_to_index_map[goal_node])

        # if it is not possible to find a path
        if len(paths) == 0:
            return None

        path = paths[0]

        # build the path
        result = [start]

        for source, target in zip(path, path[1:]):

            intersection: BoundingBox = self.graph.get_edge_data(source, target)
            x_target = intersection.x_interval.center()
            y_target = intersection.y_interval.center()
            z_target = intersection.z_interval.center()
            result.append(Point3(x_target, y_target, z_target))

        result.append(goal)
        return result

    @classmethod
    def _make_search_space(cls, search_space: Optional[BoundingBoxCollection] = None):
        """
        Create the default search space if it is not given.
        """
        if search_space is None:
            search_space = BoundingBox(-np.inf, -np.inf, -np.inf, np.inf, np.inf, np.inf).as_collection()
        return search_space

    @classmethod
    def obstacles_of_world(cls, world: World, search_space: Optional[BoundingBoxCollection] = None,
                           bloat_obstacles: float = 0.) -> Event:
        """
        Get all obstacles of the world besides the robot as a random event.

        :param world: The world to get the obstacles from.
        :param search_space: The search space for the connectivity graph.
        :param bloat_obstacles: The amount to bloat the obstacles.

        :return: An event representing the obstacles in the search space.
        """
        world_view = EnvironmentView(root=world.root)
        return cls.obstacles_from_views(world_view, search_space=search_space, bloat_obstacles=bloat_obstacles)


    @classmethod
    def obstacles_from_views(cls, obstacle_view: View, wall_view: Optional[View] = None, search_space: Optional[BoundingBoxCollection] = None,
                             bloat_obstacles: float = 0., bloat_walls: float = 0., keep_z = True) -> Event:

        """
        Create a connectivity graph from a list of views.

        :param obstacle_view: The view to create the connectivity graph from.
        :param wall_view: An optional view containing walls to be considered as obstacles.
        :param search_space: The search space for the connectivity graph.
        :param bloat_obstacles: The amount to bloat the obstacles.
        :param bloat_walls: The amount to bloat the walls.
        :param keep_z: If True, the z-axis is kept in the resulting event. Default is True.

        :return: An event representing the obstacles in the search space.
        """
        def bloat_obstacle(bb):
            return bb.bloat(bloat_obstacles, bloat_obstacles, 0.01)

        def bloat_wall(bb):
            if bb.width > bb.depth:
                return bb.bloat(bloat_walls, 0, 0.01)
            else:
                return bb.bloat(0, bloat_walls, 0.01)

        bloated_obstacles: BoundingBoxCollection = BoundingBoxCollection([
            bloat_obstacle(bb) for bb in obstacle_view.as_bounding_box_collection()
        ])

        if wall_view is not None:
            bloated_walls: BoundingBoxCollection = BoundingBoxCollection([
                bloat_wall(bb) for bb in wall_view.as_bounding_box_collection()
            ])
            bloated_obstacles.merge(bloated_walls)


        search_space = cls._make_search_space(search_space)
        return cls.obstacles_from_bounding_boxes(bloated_obstacles, search_space.event, keep_z)

    @classmethod
    def obstacles_from_bounding_boxes(cls, bounding_boxes: BoundingBoxCollection, search_space_event: Event,
                                      keep_z: bool = True) -> Optional[Event]:
        """
        Create a connectivity graph from a list of bounding boxes.

        :param bounding_boxes: The list of bounding boxes to create the connectivity graph from.
        :param search_space_event: The search space event to limit the connectivity graph to.
        :param keep_z: If True, the z-axis is kept in the resulting event. Default is True.

        :return: An event representing the obstacles in the search space, or None if no obstacles are found.
        """

        if not keep_z:
            search_space_event = search_space_event.marginal(SpatialVariables.xy)

        events = (
            bb.simple_event.as_composite_set() & search_space_event
            for bb in bounding_boxes
        )

        # skip bbs outside the search space
        events = (event for event in events if not event.is_empty())

        if not keep_z:
            events = (event.marginal(SpatialVariables.xy) for event in events)

        try:
            return reduce(or_, events)
        except TypeError:
            logger.warning("No obstacles found in the given views. Returning None.")
            return None

    @classmethod
    def free_space_from_view(cls, obstacle_view: View, wall_view: Optional[View] = None, tolerance=.001,
                             search_space: Optional[BoundingBoxCollection] = None, bloat_obstacles: float = 0.,
                             bloat_walls: float = 0.) -> Self:
        """
        Create a connectivity graph from the free space in the belief state of the robot.

        :param obstacle_view: The view containing the obstacles.
        :param wall_view: An optional view containing walls to be considered as obstacles.
        :param tolerance: The tolerance for the intersection when calculating the connectivity.
        :param search_space: The search space for the connectivity graph.
        :param bloat_obstacles: The amount to bloat the obstacles.
        :param bloat_walls: The amount to bloat the walls.

        :return: The connectivity graph. If no obstacles are found, an empty graph is returned.
        """

        # create search space for calculations
        search_space = cls._make_search_space(search_space)

        # get obstacles
        obstacles = cls.obstacles_from_views(obstacle_view, wall_view, search_space, bloat_obstacles=bloat_obstacles,
                                             bloat_walls=bloat_walls)

        if obstacles is None or obstacles.is_empty():
            return cls(search_space=search_space)

        search_event = search_space.event

        start_time = time.time_ns()
        # calculate the free space and limit it to the searching space
        free_space = ~obstacles & search_event
        logger.info(f"Free space calculated in {(time.time_ns() - start_time) / 1e6} ms")

        # create a connectivity graph from the free space and calculate the edges
        result = cls(search_space=search_space)
        [result.add_node(bb) for bb in BoundingBoxCollection.from_event(free_space)]

        start_time = time.time_ns()
        result.calculate_connectivity(tolerance)
        logger.info(f"Connectivity calculated in {(time.time_ns() - start_time) / 1e6} ms")

        return result

    @classmethod
    def free_space_from_world(cls, world: World, tolerance=.001, search_space: Optional[BoundingBoxCollection] = None,
                              bloat_obstacles: float = 0.) -> Self:
        """
        Create a connectivity graph from the free space in the belief state of the robot.

        :param world: The belief state.
        :param tolerance: The tolerance for the intersection when calculating the connectivity.
        :param search_space: The search space for the connectivity graph.
        :param bloat_obstacles: The amount to bloat the obstacles.

        :return: The connectivity graph.
        """

        view = EnvironmentView(root=world.root, _world=world)

        return cls.free_space_from_view(view, tolerance=tolerance, search_space=search_space,
                                        bloat_obstacles=bloat_obstacles)

    @classmethod
    def navigation_map_from_view(cls, obstacle_view: View, wall_view: Optional[View] = None, tolerance=.001,
                                  search_space: Optional[BoundingBoxCollection] = None,
                                  bloat_obstacles: float = 0., bloat_walls: float = 0.) -> Self:
        """
        Create a GCS from the free space in the belief state of the robot for navigation.
        The resulting GCS describes the paths for navigation, meaning that changing the z-axis position is not
        possible.
        Furthermore, it is taken into account that the robot has to fit through the entire space and not just
        through the floor level obstacles.

        :param obstacle_view: The view containing the obstacles.
        :param wall_view: An optional view containing walls to be considered as obstacles.
        :param tolerance: The tolerance for the intersection when calculating the connectivity.
        :param search_space: The search space for the connectivity graph.
        :param bloat_obstacles: The amount to bloat the obstacles.
        :param bloat_walls: The amount to bloat the walls.

        :return: The connectivity graph. If no obstacles are found, an empty graph is returned.
        """

        # create search space for calculations
        search_space = cls._make_search_space(search_space)

        obstacles = cls.obstacles_from_views(obstacle_view, wall_view, search_space, bloat_obstacles, bloat_walls,
                                             keep_z=False)

        if obstacles is None or obstacles.is_empty():
            return cls(search_space=search_space)

        # remove the z axis
        og_search_event = search_space.event
        search_event = og_search_event.marginal(SpatialVariables.xy)

        free_space = ~obstacles & search_event

        SimpleEvent({SpatialVariables.z.value: reals()})
        # create floor level
        z_event = SimpleEvent({SpatialVariables.z.value: reals()}).as_composite_set()
        z_event.fill_missing_variables(SpatialVariables.xy)
        free_space.fill_missing_variables(SortedSet([SpatialVariables.z.value]))
        free_space &= z_event
        free_space &= og_search_event

        # create a connectivity graph from the free space and calculate the edges
        result = cls(search_space=search_space)
        free_space_boxes = BoundingBoxCollection.from_event(free_space)
        [result.add_node(bb) for bb in free_space_boxes]
        result.calculate_connectivity(tolerance)

        return result

    @classmethod
    def navigation_map_from_world(cls, world: World, tolerance=.001,
                                  search_space: Optional[BoundingBoxCollection] = None,
                                  bloat_obstacles: float = 0.) -> Self:
        """
        Create a GCS from the free space in the belief state of the robot for navigation.
        The resulting GCS describes the paths for navigation, meaning that changing the z-axis position is not
        possible.
        Furthermore, it is taken into account that the robot has to fit through the entire space and not just
        through the floor level obstacles.

        :param world: The belief state.
        :param tolerance: The tolerance for the intersection when calculating the connectivity.
        :param search_space: The search space for the connectivity graph.
        :param bloat_obstacles: The amount to bloat the obstacles.

        :return: The connectivity graph.
        """

        view = EnvironmentView(root=world.root, _world=world)

        return cls.navigation_map_from_view(view, tolerance=tolerance, search_space=search_space,
                                            bloat_obstacles=bloat_obstacles)


from typing import Tuple, Union, List

import numpy as np

from .derivatives import Derivatives
from ..types import NpMatrix4x4, AnyMatrix4x4

def quaternion_multiply(quaternion1: np.ndarray, quaternion0: np.ndarray) -> np.ndarray:
    x0, y0, z0, w0 = quaternion0
    x1, y1, z1, w1 = quaternion1
    return np.array((
        x1 * w0 + y1 * z0 - z1 * y0 + w1 * x0,
        -x1 * z0 + y1 * w0 + z1 * x0 + w1 * y0,
        x1 * y0 - y1 * x0 + z1 * w0 + w1 * z0,
        -x1 * x0 - y1 * y0 - z1 * z0 + w1 * w0), dtype=np.float64)


def quaternion_conjugate(quaternion: np.ndarray) -> np.ndarray:
    return np.array((-quaternion[0], -quaternion[1],
                     -quaternion[2], quaternion[3]), dtype=np.float64)


def qv_mult(quaternion: np.ndarray, vector: np.ndarray) -> np.ndarray:
    """
    Transforms a vector by a quaternion
    :param quaternion: Quaternion
    :param vector: vector
    :return: transformed vector
    """
    q = quaternion
    v = [vector[0], vector[1], vector[2], 0]
    return quaternion_multiply(quaternion_multiply(q, v), quaternion_conjugate(q))[:-1]


def quaternion_from_axis_angle(axis: Union[List[float], Tuple[float, float, float], np.ndarray],
                               angle: float) -> np.ndarray:
    half_angle = angle / 2
    return np.array([axis[0] * np.sin(half_angle),
                     axis[1] * np.sin(half_angle),
                     axis[2] * np.sin(half_angle),
                     np.cos(half_angle)], dtype=np.float64)


_EPS = np.finfo(float).eps * 4.0


def rpy_from_matrix(rotation_matrix: NpMatrix4x4) -> Tuple[float, float, float]:
    """
    :param rotation_matrix: 4x4 Matrix
    :type rotation_matrix: Matrix
    :return: roll, pitch, yaw
    :rtype: (Union[float, Symbol], Union[float, Symbol], Union[float, Symbol])
    """
    i = 0
    j = 1
    k = 2

    cy = np.sqrt(rotation_matrix[i, i] * rotation_matrix[i, i] + rotation_matrix[j, i] * rotation_matrix[j, i])
    if cy - _EPS > 0:
        roll = np.arctan2(rotation_matrix[k, j], rotation_matrix[k, k])
        pitch = np.arctan2(-rotation_matrix[k, i], cy)
        yaw = np.arctan2(rotation_matrix[j, i], rotation_matrix[i, i])
    else:
        roll = np.arctan2(-rotation_matrix[j, k], rotation_matrix[j, j])
        pitch = np.arctan2(-rotation_matrix[k, i], cy)
        yaw = 0
    return roll, pitch, yaw


def rpy_from_quaternion(qx: float, qy: float, qz: float, qw: float) -> Tuple[float, float, float]:
    return rpy_from_matrix(rotation_matrix_from_quaternion(qx, qy, qz, qw))


def rotation_matrix_from_rpy(roll: float, pitch: float, yaw: float) -> NpMatrix4x4:
    """
    Conversion of roll, pitch, yaw to 4x4 rotation matrix according to:
    https://github.com/orocos/orocos_kinematics_dynamics/blob/master/orocos_kdl/src/frames.cpp#L167
    :return: 4x4 Matrix
    """
    rx = np.array([[1, 0, 0, 0],
                   [0, np.cos(roll), -np.sin(roll), 0],
                   [0, np.sin(roll), np.cos(roll), 0],
                   [0, 0, 0, 1]])
    ry = np.array([[np.cos(pitch), 0, np.sin(pitch), 0],
                   [0, 1, 0, 0],
                   [-np.sin(pitch), 0, np.cos(pitch), 0],
                   [0, 0, 0, 1]])
    rz = np.array([[np.cos(yaw), -np.sin(yaw), 0, 0],
                   [np.sin(yaw), np.cos(yaw), 0, 0],
                   [0, 0, 1, 0],
                   [0, 0, 0, 1]])
    return np.dot(np.dot(rz, ry), rx)


def rotation_matrix_from_quaternion(x: float, y: float, z: float, w: float) -> NpMatrix4x4:
    """
    Unit quaternion to 4x4 rotation matrix according to:
    https://github.com/orocos/orocos_kinematics_dynamics/blob/master/orocos_kdl/src/frames.cpp#L167
    :return: 4x4 Matrix
    """
    x2 = x * x
    y2 = y * y
    z2 = z * z
    w2 = w * w
    return np.array([[w2 + x2 - y2 - z2, 2 * x * y - 2 * w * z, 2 * x * z + 2 * w * y, 0],
                     [2 * x * y + 2 * w * z, w2 - x2 + y2 - z2, 2 * y * z - 2 * w * x, 0],
                     [2 * x * z - 2 * w * y, 2 * y * z + 2 * w * x, w2 - x2 - y2 + z2, 0],
                     [0, 0, 0, 1]], dtype=np.float64)


def rotation_matrix_from_axis_angle(axis: Union[List[float], Tuple[float, float, float], np.ndarray], angle: float) \
        -> NpMatrix4x4:
    return rotation_matrix_from_quaternion(*quaternion_from_axis_angle(axis, angle))


def quaternion_from_rpy(roll: float, pitch: float, yaw: float) -> np.ndarray:
    return quaternion_from_rotation_matrix(rotation_matrix_from_rpy(roll, pitch, yaw))


def quaternion_from_rotation_matrix(matrix: AnyMatrix4x4) -> np.ndarray:
    """
    :param matrix: 4x4 Matrix
    :return: array length 4
    """
    q = np.empty((4,), dtype=np.float64)
    M = np.array(matrix, dtype=np.float64)[:4, :4]
    t = np.trace(M)
    if t > M[3, 3]:
        q[3] = t
        q[2] = M[1, 0] - M[0, 1]
        q[1] = M[0, 2] - M[2, 0]
        q[0] = M[2, 1] - M[1, 2]
    else:
        i, j, k = 0, 1, 2
        if M[1, 1] > M[0, 0]:
            i, j, k = 1, 2, 0
        if M[2, 2] > M[i, i]:
            i, j, k = 2, 0, 1
        t = M[i, i] - (M[j, j] + M[k, k]) + M[3, 3]
        q[i] = t
        q[j] = M[i, j] + M[j, i]
        q[k] = M[k, i] + M[i, k]
        q[3] = M[k, j] - M[j, k]
    q *= 0.5 / np.sqrt(t * M[3, 3])
    return q


def axis_angle_from_quaternion(x: float, y: float, z: float, w: float) -> Tuple[np.ndarray, float]:
    l = np.linalg.norm(np.array([x, y, z, w]))
    x, y, z, w = x / l, y / l, z / l, w / l
    w2 = np.sqrt(1 - w ** 2)
    if w2 == 0:
        m = 1
    else:
        m = w2
    if w2 == 0:
        angle = 0
    else:
        angle = (2 * np.arccos(min(max(-1, w), 1)))
    if w2 == 0:
        x = 0
        y = 0
        z = 1
    else:
        x = x / m
        y = y / m
        z = z / m
    return np.array([x, y, z]), angle


def axis_angle_from_rotation_matrix(m: NpMatrix4x4) -> Tuple[np.ndarray, float]:
    return axis_angle_from_quaternion(*quaternion_from_rotation_matrix(m))


def axis_angle_from_rpy(roll: float, pitch: float, yaw: float) -> Tuple[np.ndarray, float]:
    return axis_angle_from_quaternion(*quaternion_from_rpy(roll, pitch, yaw))


def gauss(n: float) -> float:
    return (n ** 2 + n) / 2


def max_velocity_from_horizon_and_jerk(prediction_horizon: int,
                                       vel_limit: float,
                                       acc_limit: float,
                                       jerk_limit: float,
                                       sample_period: float,
                                       max_derivative: Derivatives):
    n2 = int((prediction_horizon) / 2)
    vel_jerk_limited = (gauss(n2) + gauss(n2 - 1)) * jerk_limit * sample_period ** 2
    vel_acc_limits = acc_limit * sample_period * prediction_horizon
    if max_derivative == Derivatives.jerk:
        return min(vel_limit, vel_acc_limits, vel_jerk_limited)
    if max_derivative == Derivatives.acceleration:
        return min(vel_limit, vel_acc_limits)
    return vel_limit


def limit(a: float, lower_limit: float, upper_limit: float) -> float:
    return max(lower_limit, min(upper_limit, a))


def inverse_frame(f1_T_f2: NpMatrix4x4) -> NpMatrix4x4:
    """
    :param f1_T_f2: 4x4 Matrix
    :return: f2_T_f1
    """
    R = f1_T_f2[:3, :3]
    t = f1_T_f2[:3, 3]
    Rt = R.T
    f2_T_f1 = np.empty((4, 4), dtype=f1_T_f2.dtype)
    f2_T_f1[:3, :3] = Rt
    f2_T_f1[:3, 3] = -Rt @ t
    f2_T_f1[3] = [0, 0, 0, 1]
    return f2_T_f1


def angle_between_vector(v1: np.ndarray, v2: np.ndarray) -> float:
    """
    :param v1: vector length 3
    :param v2: vector length 3
    """
    return np.arccos(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)))


def normalize(v: np.ndarray) -> np.ndarray:
    return v / np.linalg.norm(v)


def my_cross(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
    """
    :param v1: vector length 3 or 4
    :param v2: vector length 3 or 4
    :return: cross product vector length 3
    """
    return np.cross(v1[:-1], v2[:-1])


def point_to_single_caster_angle(px, py, caster_v, forward_velocity):
    max_angular_velocity = 1
    z = np.array([0, 0, 1, 0])
    x = np.array([1, 0, 0, 0])
    center_P_p = np.array([px, py, 0, 1])
    c_V_p = center_P_p - caster_v
    c_V_goal = my_cross(c_V_p, z)
    angle = angle_between_vector(x[:-1], c_V_goal)
    radius = np.linalg.norm(c_V_p)
    if radius > 0.01:
        circumference = 2 * np.pi * radius
        number_of_revolutions = forward_velocity / circumference
        angular_velocity = number_of_revolutions / (2 * np.pi)
        angular_velocity = min(max(angular_velocity, -max_angular_velocity), max_angular_velocity)
    else:
        angular_velocity = max_angular_velocity
    print(f'angular velocity {angular_velocity}')
    return angle, c_V_goal


def point_to_caster_angles(px, py):
    forward_velocity = 2
    # center_P_fl = np.array([1, 1, 0, 1])
    # print(f'fl {point_to_single_caster_angle(px, py, center_P_fl, forward_velocity)}')
    # center_P_fr = np.array([1, -1, 0, 1])
    # print(f'fr {point_to_single_caster_angle(px, py, center_P_fr, forward_velocity)}')
    # center_P_bl = np.array([-1, 1, 0, 1])
    # print(f'bl {point_to_single_caster_angle(px, py, center_P_bl, forward_velocity)}')
    # center_P_br = np.array([-1, -1, 0, 1])
    # print(f'br {point_to_single_caster_angle(px, py, center_P_br, forward_velocity)}')
    center = np.array([0, 0, 0, 1])
    print(f'center {point_to_single_caster_angle(px, py, center, forward_velocity)}')


def shortest_angular_distance(from_angle, to_angle):
    """ Given 2 angles, this returns the shortest angular
        difference.  The inputs and ouputs are of course radians.

        The result would always be -pi <= result <= pi. Adding the result
        to "from" will always get you an equivelent angle to "to".
    """
    return normalize_angle(to_angle - from_angle)


def normalize_angle(angle):
    """ Normalizes the angle to be -pi to +pi
        It takes and returns radians."""
    a = normalize_angle_positive(angle)
    if a > np.pi:
        a -= 2.0 * np.pi
    return a


def normalize_angle_positive(angle):
    """ Normalizes the angle to be 0 to 2*pi
        It takes and returns radians. """
    return angle % (2.0 * np.pi)


def quaternion_slerp(q1, q2, t):
    """
    spherical linear interpolation that takes into account that q == -q
    :param q1: 4x1 Matrix
    :param q2: 4x1 Matrix
    :param t: float, 0-1
    :return: 4x1 Matrix; Return spherical linear interpolation between two quaternions.
    """
    cos_half_theta = q1.dot(q2)

    if cos_half_theta < 0:
        q2 = -q2
        cos_half_theta = -cos_half_theta

    if abs(cos_half_theta) > 1:
        return q1
    # enforce acos(x) with -1 < x < 1
    cos_half_theta = min(1, cos_half_theta)
    cos_half_theta = max(-1, cos_half_theta)

    half_theta = np.arccos(cos_half_theta)

    sin_half_theta = np.sqrt(1.0 - cos_half_theta * cos_half_theta)

    ratio_a = np.sin((1.0 - t) * half_theta) / sin_half_theta
    ratio_b = np.sin(t * half_theta) / sin_half_theta

    if 0.001 > abs(sin_half_theta):
        return 0.5 * q1 + 0.5 * q2
    return ratio_a * q1 + ratio_b * q2




import os
from dataclasses import dataclass
from typing import Optional

import numpy
from multiverse_parser import (InertiaSource,
                               UsdImporter, MjcfImporter, UrdfImporter,
                               BodyBuilder,
                               JointBuilder, JointType)
from pxr import UsdUrdf

from ..connections import RevoluteConnection, PrismaticConnection, FixedConnection
from ..spatial_types.derivatives import DerivativeMap
from ..prefixed_name import PrefixedName
from ..spatial_types import spatial_types as cas
from ..world import World, Body, Connection


@dataclass
class MultiParser:
    """
    Class to parse any scene description files to worlds.
    """

    file_path: str
    """
    The file path of the scene.
    """

    prefix: Optional[str] = None
    """
    The prefix for every name used in this world.
    """

    def __post_init__(self):
        if self.prefix is None:
            self.prefix = os.path.basename(self.file_path).split('.')[0]

    def parse(self) -> World:
        fixed_base = True
        root_name = None
        with_physics = True
        with_visual = True
        with_collision = True
        inertia_source = InertiaSource.FROM_SRC
        default_rgba = numpy.array([0.9, 0.9, 0.9, 1.0])

        file_ext = os.path.splitext(self.file_path)[1]
        if file_ext in [".usd", ".usda", ".usdc"]:
            add_xform_for_each_geom = True
            factory = UsdImporter(file_path=self.file_path,
                                  fixed_base=fixed_base,
                                  root_name=root_name,
                                  with_physics=with_physics,
                                  with_visual=with_visual,
                                  with_collision=with_collision,
                                  inertia_source=inertia_source,
                                  default_rgba=default_rgba,
                                  add_xform_for_each_geom=add_xform_for_each_geom)
        elif file_ext == ".urdf":
            factory = UrdfImporter(file_path=self.file_path,
                                   fixed_base=fixed_base,
                                   root_name=root_name,
                                   with_physics=with_physics,
                                   with_visual=with_visual,
                                   with_collision=with_collision,
                                   inertia_source=inertia_source,
                                   default_rgba=default_rgba)
        elif file_ext == ".xml":
            if root_name is None:
                root_name = "world"
            factory = MjcfImporter(file_path=self.file_path,
                                   fixed_base=fixed_base,
                                   root_name=root_name,
                                   with_physics=with_physics,
                                   with_visual=with_visual,
                                   with_collision=with_collision,
                                   inertia_source=inertia_source,
                                   default_rgba=default_rgba)
        else:
            raise NotImplementedError(f"Importing from {file_ext} is not supported yet.")

        factory.import_model()
        bodies = [self.parse_body(body_builder) for body_builder in factory.world_builder.body_builders]
        world = World()
        world.add_body(bodies[0])

        with world.modify_world():
            for body in bodies:
                world.add_body(body)
            joints = []
            for body_builder in factory.world_builder.body_builders:
                joints += self.parse_joints(body_builder=body_builder, world=world)
            for joint in joints:
                world.add_connection(joint)

        return world

    def parse_joints(self, body_builder: BodyBuilder, world: World) -> list[Connection]:
        """
        Parses joints from a BodyBuilder instance.
        :param body_builder: The BodyBuilder instance to parse.
        :param world: The World instance to add the connections to.
        :return: A list of Connection instances representing the parsed joints.
        """
        connections = []
        for joint_builder in body_builder.joint_builders:
            parent_body = world.get_body_by_name(joint_builder.parent_prim.GetName())
            child_body = world.get_body_by_name(joint_builder.child_prim.GetName())
            connection = self.parse_joint(joint_builder, parent_body, child_body, world)
            connections.append(connection)
        if len(body_builder.joint_builders) == 0 and not body_builder.xform.GetPrim().GetParent().IsPseudoRoot():
            parent_body = world.get_body_by_name(body_builder.xform.GetPrim().GetParent().GetName())
            child_body = world.get_body_by_name(body_builder.xform.GetPrim().GetName())
            transform = body_builder.xform.GetLocalTransformation()
            pos = transform.ExtractTranslation()
            quat = transform.ExtractRotationQuat()
            point_expr = cas.Point3(pos[0], pos[1], pos[2])
            quaternion_expr = cas.Quaternion(quat.GetImaginary()[0],
                                              quat.GetImaginary()[1],
                                              quat.GetImaginary()[2],
                                              quat.GetReal())
            origin = cas.TransformationMatrix.from_point_rotation_matrix(point=point_expr,
                                                                         rotation_matrix=quaternion_expr.to_rotation_matrix())
            connection = FixedConnection(parent=parent_body, child=child_body, origin_expression=origin)
            connections.append(connection)

        return connections

    def parse_joint(self, joint_builder: JointBuilder, parent_body: Body, child_body: Body, world: World) -> Connection:
        joint_prim = joint_builder.joint.GetPrim()
        joint_name = joint_prim.GetName()
        joint_pos = joint_builder.pos
        joint_quat = joint_builder.quat
        point_expr = cas.Point3(joint_pos[0], joint_pos[1], joint_pos[2])
        quaternion_expr = cas.Quaternion(joint_quat.GetImaginary()[0],
                                          joint_quat.GetImaginary()[1],
                                          joint_quat.GetImaginary()[2],
                                          joint_quat.GetReal())
        origin = cas.TransformationMatrix.from_point_rotation_matrix(point=point_expr,
                                                                     rotation_matrix=quaternion_expr.to_rotation_matrix())
        free_variable_name = PrefixedName(joint_name)
        offset = None
        multiplier = None
        if joint_prim.HasAPI(UsdUrdf.UrdfJointAPI):
            urdf_joint_api = UsdUrdf.UrdfJointAPI(joint_prim)
            if len(urdf_joint_api.GetJointRel().GetTargets()) > 0:
                free_variable_name = PrefixedName(urdf_joint_api.GetJointRel().GetTargets()[0].name)
                offset = urdf_joint_api.GetOffsetAttr().Get()
                multiplier = urdf_joint_api.GetMultiplierAttr().Get()
        if joint_builder.type == JointType.FREE:
            raise NotImplementedError("Free joints are not supported yet.")
        elif joint_builder.type == JointType.FIXED:
            return FixedConnection(parent=parent_body, child=child_body, origin_expression=origin)
        elif joint_builder.type in [JointType.REVOLUTE, JointType.CONTINUOUS, JointType.PRISMATIC]:
            axis = cas.Vector3(float(joint_builder.axis.to_array()[0]),
                                   float(joint_builder.axis.to_array()[1]),
                                   float(joint_builder.axis.to_array()[2]),
                                   reference_frame=parent_body)
            try:
                dof = world.get_degree_of_freedom_by_name(free_variable_name)
            except KeyError:
                if joint_builder.type == JointType.CONTINUOUS:
                    lower_limits = DerivativeMap()
                    lower_limits.position = joint_builder.joint.GetLowerLimitAttr().Get()
                    upper_limits = DerivativeMap()
                    upper_limits.position = joint_builder.joint.GetUpperLimitAttr().Get()
                    dof = world.create_degree_of_freedom(name=PrefixedName(joint_name),
                                                         lower_limits=lower_limits,
                                                         upper_limits=upper_limits)
                else:
                    dof = world.create_degree_of_freedom(name=PrefixedName(joint_name))
            if joint_builder.type in [JointType.REVOLUTE, JointType.CONTINUOUS]:
                connection = RevoluteConnection(parent=parent_body, child=child_body, origin_expression=origin,
                                                multiplier=multiplier, offset=offset,
                                                axis=axis, dof=dof)
            else:
                connection = PrismaticConnection(parent=parent_body, child=child_body, origin_expression=origin,
                                                 multiplier=multiplier, offset=offset,
                                                 axis=axis, dof=dof)
            return connection
        else:
            raise NotImplementedError(f"Joint type {joint_builder.type} is not supported yet.")

    def parse_body(self, body_builder: BodyBuilder) -> Body:
        """
        Parses a body from a BodyBuilder instance.
        :param body_builder: The BodyBuilder instance to parse.
        :return: A Body instance representing the parsed body.
        """
        name = PrefixedName(prefix=self.prefix, name=body_builder.xform.GetPrim().GetName())
        return Body(name=name)



import os
import unittest

import plotly.graph_objects as go
from random_events.interval import SimpleInterval
from random_events.product_algebra import SimpleEvent

from semantic_world.adapters.urdf import URDFParser
from semantic_world.geometry import BoundingBox, BoundingBoxCollection
from semantic_world.graph_of_convex_sets import GraphOfConvexSets, PoseOccupiedError
from semantic_world.spatial_types import Point3
from semantic_world.variables import SpatialVariables
from semantic_world.world import World


class GCSTestCase(unittest.TestCase):
    """
    Testcase to test the navigation around a unit box.
    """

    gcs: GraphOfConvexSets

    @classmethod
    def setUpClass(cls):
        gcs = GraphOfConvexSets()

        obstacle = BoundingBox(0, 0, 0, 1, 1, 1)

        z_lim = SimpleInterval(.45, .55)
        x_lim = SimpleInterval(-2, 3)
        y_lim = SimpleInterval(-2, 3)
        limiting_event = SimpleEvent({SpatialVariables.x.value: x_lim,
                                      SpatialVariables.y.value: y_lim,
                                      SpatialVariables.z.value: z_lim,})
        obstacles = BoundingBoxCollection.from_event(
            ~obstacle.simple_event.as_composite_set() & limiting_event.as_composite_set())
        [gcs.add_node(bb) for bb in obstacles]
        gcs.calculate_connectivity()
        cls.gcs = gcs

    def test_reachability(self):
        start_point = Point3(-1, -1, 0.5)
        target_point = Point3(2, 2, 0.5)

        path = self.gcs.path_from_to(start_point, target_point)
        self.assertEqual(len(path), 4)

    def test_plot(self):
        free_space_plot = go.Figure(self.gcs.plot_free_space())
        self.assertIsNotNone(free_space_plot)
        occupied_space_plot = go.Figure(self.gcs.plot_occupied_space())
        self.assertIsNotNone(occupied_space_plot)


class GCSFromWorldTestCase(unittest.TestCase):
    """
    Test the application of a connectivity graph to the belief state.
    """

    world: World

    @classmethod
    def setUpClass(cls):
        urdf_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "resources", "urdf")
        apartment = os.path.join(urdf_dir, "table.urdf")
        apartment_parser = URDFParser(apartment)
        cls.world = apartment_parser.parse()

    def test_from_world(self):
        search_space = BoundingBox(min_x=-5, max_x=-2,
                                   min_y=-1, max_y=2,
                                   min_z=0, max_z=2).as_collection()
        gcs = GraphOfConvexSets.free_space_from_world(self.world, search_space=search_space)
        self.assertIsNotNone(gcs)
        self.assertGreater(len(gcs.graph.nodes()), 0)
        self.assertGreater(len(gcs.graph.edges()), 0)

        start = Point3(-4.5, -0.5, 0.4)
        target = Point3(-2.5, 1.5, 0.9)

        path = gcs.path_from_to(start, target)

        self.assertIsNotNone(path)
        self.assertGreater(len(path), 1)

        with self.assertRaises(PoseOccupiedError):
            start = Point3(-10, -10, -10)
            target = Point3(10, 10, 10)
            gcs.path_from_to(start, target)

    def test_navigation_map_from_world(self):
        search_space = BoundingBox(min_x=-5, max_x=-2,
                                   min_y=-1, max_y=2,
                                   min_z=0, max_z=2).as_collection()
        gcs = GraphOfConvexSets.navigation_map_from_world(self.world, search_space=search_space)
        self.assertGreater(len(gcs.graph.nodes()), 0)
        self.assertGreater(len(gcs.graph.edges()), 0)


if __name__ == '__main__':
    unittest.main()


from __future__ import annotations

import os
from copy import deepcopy
from functools import lru_cache, wraps
from typing import Any, Tuple
from xml.etree import ElementTree as ET


class IDGenerator:
    """
    A class that generates incrementing, unique IDs and caches them for every object this is called on.
    """

    _counter = 0
    """
    The counter of the unique IDs.
    """

    @lru_cache(maxsize=None)
    def __call__(self, obj: Any) -> int:
        """
        Creates a unique ID and caches it for every object this is called on.

        :param obj: The object to generate a unique ID for, must be hashable.
        :return: The unique ID.
        """
        self._counter += 1
        return self._counter


class suppress_stdout_stderr(object):
    """
    A context manager for doing a "deep suppression" of stdout and stderr in
    Python, i.e. will suppress all prints, even if the print originates in a
    compiled C/Fortran sub-function.

    This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).
    Copied from https://stackoverflow.com/questions/11130156/suppress-stdout-stderr-print-from-python-functions
    """

    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for _ in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        # This one is not needed for URDF parsing output
        # os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        # This one is not needed for URDF parsing output
        # os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)
        # Close all file descriptors
        for fd in self.null_fds + self.save_fds:
            os.close(fd)


def hacky_urdf_parser_fix(urdf: str, blacklist: Tuple[str] = ('transmission', 'gazebo')) -> str:
    # Parse input string
    root = ET.fromstring(urdf)

    # Iterate through each section in the blacklist
    for section_name in blacklist:
        # Find all sections with the given name and remove them
        for elem in root.findall(f".//{section_name}"):
            parent = root.find(f".//{section_name}/..")
            if parent is not None:
                parent.remove(elem)

    # Turn back to string
    return ET.tostring(root, encoding='unicode')


def robot_name_from_urdf_string(urdf_string: str) -> str:
    """
    Returns the name defined in the robot tag, e.g., 'pr2' from <robot name="pr2"> ... </robot>.
    :param urdf_string: URDF string
    :return: Extracted name
    """
    return urdf_string.split('robot name="')[1].split('"')[0]


def copy_lru_cache(maxsize=None, typed=False):
    def decorator(func):
        cached_func = lru_cache(maxsize=maxsize, typed=typed)(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            result = cached_func(*args, **kwargs)
            return deepcopy(result)

        # Preserve lru_cache methods
        wrapper.cache_info = cached_func.cache_info
        wrapper.cache_clear = cached_func.cache_clear

        return wrapper

    return decorator



from __future__ import annotations

import itertools
from abc import ABC
from dataclasses import dataclass, field
from functools import cached_property
from typing import Optional, List, Iterator

from typing_extensions import Self

import trimesh
from random_events.interval import SimpleInterval, Bound
from random_events.product_algebra import SimpleEvent, Event

from .spatial_types import TransformationMatrix, Point3
from .spatial_types.spatial_types import Expression
from .utils import IDGenerator
from .variables import SpatialVariables

id_generator = IDGenerator()

@dataclass
class Color:
    """
    Dataclass for storing rgba_color as an RGBA value.
    The values are stored as floats between 0 and 1.
    The default rgba_color is white.
    """
    R: float = 1
    """
    Red value of the color.
    """

    G: float = 1
    """
    Green value of the color.
    """

    B: float = 1
    """
    Blue value of the color.
    """

    A: float = 1
    """
    Opacity of the color.
    """

@dataclass
class Scale:
    """
    Dataclass for storing the scale of geometric objects.
    """

    x: float = 1.
    """
    The scale in the x direction.
    """

    y: float = 1.
    """
    The scale in the y direction.
    """

    z: float = 1.
    """
    The scale in the z direction.
    """

@dataclass
class Shape(ABC):
    """
    Base class for all shapes in the world.
    """
    origin: TransformationMatrix = field(default_factory=TransformationMatrix)

    def as_bounding_box(self) -> BoundingBox:
        """
        Returns the bounding box of the shape.
        This method should be implemented by subclasses.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    @property
    def mesh(self) -> trimesh.Trimesh:
        """
        The mesh object of the shape.
        This should be implemented by subclasses.
        """
        raise NotImplementedError("Subclasses must implement the mesh property.")

@dataclass
class Mesh(Shape):
    """
    A mesh shape.
    """

    filename: str = ""
    """
    Filename of the mesh.
    """

    scale: Scale = field(default_factory=Scale)
    """
    Scale of the mesh.
    """

    @cached_property
    def mesh(self) -> trimesh.Trimesh:
        """
        The mesh object.
        """
        return trimesh.load_mesh(self.filename)

    def as_bounding_box(self) -> BoundingBox:
        """
        Returns the bounding box of the mesh.
        """
        return BoundingBox.from_mesh(self.mesh)




@dataclass
class Primitive(Shape):
    """
    A primitive shape.
    """
    color: Color = field(default_factory=Color)


@dataclass
class Sphere(Primitive):
    """
    A sphere shape.
    """

    radius: float = 0.5
    """
    Radius of the sphere.
    """

    @property
    def mesh(self) -> trimesh.Trimesh:
        """
        Returns a trimesh object representing the sphere.
        """
        return trimesh.creation.icosphere(subdivisions=2, radius=self.radius)

    def as_bounding_box(self) -> BoundingBox:
        """
        Returns the bounding box of the sphere.
        """
        return BoundingBox(-self.radius, -self.radius, -self.radius,
                           self.radius, self.radius, self.radius)


@dataclass
class Cylinder(Primitive):
    """
    A cylinder shape.
    """
    width: float = 0.5
    height: float = 0.5

    @property
    def mesh(self) -> trimesh.Trimesh:
        """
        Returns a trimesh object representing the cylinder.
        """
        return trimesh.creation.cylinder(radius=self.width / 2, height=self.height, sections=16)



    def as_bounding_box(self) -> BoundingBox:
        """
        Returns the bounding box of the cylinder.
        The bounding box is axis-aligned and centered at the origin.
        """
        half_width = self.width / 2
        half_height = self.height / 2
        return BoundingBox(-half_width, -half_width, -half_height,
                           half_width, half_width, half_height)


@dataclass
class Box(Primitive):
    """
    A box shape. Pivot point is at the center of the box.
    """
    scale: Scale = field(default_factory=Scale)

    @property
    def mesh(self) -> trimesh.Trimesh:
        """
        Returns a trimesh object representing the box.
        The box is centered at the origin and has the specified scale.
        """
        return trimesh.creation.box(extents=(self.scale.x, self.scale.y, self.scale.z))

    def as_bounding_box(self) -> BoundingBox:
        """
        Returns the bounding box of the box.
        """
        return BoundingBox(-self.scale.x / 2, -self.scale.y / 2, -self.scale.z / 2,
                           self.scale.x / 2, self.scale.y / 2, self.scale.z / 2)


@dataclass
class BoundingBox:
    min_x: float
    """
    The minimum x-coordinate of the bounding box.
    """

    min_y: float
    """
    The minimum y-coordinate of the bounding box.
    """

    min_z: float
    """
    The minimum z-coordinate of the bounding box.
    """

    max_x: float
    """
    The maximum x-coordinate of the bounding box.
    """

    max_y: float
    """
    The maximum y-coordinate of the bounding box.
    """

    max_z: float
    """
    The maximum z-coordinate of the bounding box.
    """

    def __hash__(self):
        # The hash should be this since comparing those via hash is checking if those are the same and not just equal
        return hash((self.min_x, self.min_y, self.min_z, self.max_x, self.max_y, self.max_z))

    @property
    def x_interval(self) -> SimpleInterval:
        """
        :return: The x interval of the bounding box.
        """
        return SimpleInterval(self.min_x, self.max_x, Bound.CLOSED, Bound.CLOSED)

    @property
    def y_interval(self) -> SimpleInterval:
        """
        :return: The y interval of the bounding box.
        """
        return SimpleInterval(self.min_y, self.max_y, Bound.CLOSED, Bound.CLOSED)

    @property
    def z_interval(self) -> SimpleInterval:
        """
        :return: The z interval of the bounding box.
        """
        return SimpleInterval(self.min_z, self.max_z, Bound.CLOSED, Bound.CLOSED)

    @property
    def depth(self) -> float:
        return self.max_x - self.min_x

    @property
    def height(self) -> float:
        return self.max_z - self.min_z

    @property
    def width(self) -> float:
        return self.max_y - self.min_y

    @property
    def simple_event(self) -> SimpleEvent:
        """
        :return: The bounding box as a random event.
        """
        return SimpleEvent({SpatialVariables.x.value: self.x_interval,
                            SpatialVariables.y.value: self.y_interval,
                            SpatialVariables.z.value: self.z_interval})

    def bloat(self, x_amount: float = 0., y_amount: float = 0, z_amount: float = 0) -> BoundingBox:
        """
        Enlarges the bounding box by a given amount in all dimensions.

        :param x_amount: The amount to adjust minimum and maximum x-coordinates
        :param y_amount: The amount to adjust minimum and maximum y-coordinates
        :param z_amount: The amount to adjust minimum and maximum z-coordinates
        :return: New enlarged bounding box
        """
        return self.__class__(self.min_x - x_amount, self.min_y - y_amount, self.min_z - z_amount,
                              self.max_x + x_amount, self.max_y + y_amount, self.max_z + z_amount)

    def contains(self, point: Point3) -> bool:
        """
        Check if the bounding box contains a point.
        """
        x, y, z = (point.x.to_np(), point.y.to_np(), point.z.to_np()) if isinstance(point.z, Expression) else (point.x, point.y, point.z)

        return self.simple_event.contains((x, y, z))

    def as_collection(self) -> BoundingBoxCollection:
        """
        Convert the bounding box to a collection of bounding boxes.

        :return: The bounding box as a collection
        """
        return BoundingBoxCollection([self])

    @classmethod
    def from_simple_event(cls, simple_event: SimpleEvent):
        """
        Create a list of bounding boxes from a simple random event.

        :param simple_event: The random event.
        :return: The list of bounding boxes.
        """
        result = []
        for x, y, z in itertools.product(simple_event[SpatialVariables.x.value].simple_sets,
                                         simple_event[SpatialVariables.y.value].simple_sets,
                                         simple_event[SpatialVariables.z.value].simple_sets):
            result.append(cls(x.lower, y.lower, z.lower, x.upper, y.upper, z.upper))
        return result

    def intersection_with(self, other: BoundingBox) -> Optional[BoundingBox]:
        """
        Compute the intersection of two bounding boxes.

        :param other: The other bounding box.
        :return: The intersection of the two bounding boxes or None if they do not intersect.
        """
        result = self.simple_event.intersection_with(other.simple_event)
        if result.is_empty():
            return None
        return self.__class__.from_simple_event(result)[0]

    def enlarge(self, min_x: float = 0., min_y: float = 0, min_z: float = 0,
                max_x: float = 0., max_y: float = 0., max_z: float = 0.):
        """
        Enlarge the axis-aligned bounding box by a given amount in-place.
        :param min_x: The amount to enlarge the minimum x-coordinate
        :param min_y: The amount to enlarge the minimum y-coordinate
        :param min_z: The amount to enlarge the minimum z-coordinate
        :param max_x: The amount to enlarge the maximum x-coordinate
        :param max_y: The amount to enlarge the maximum y-coordinate
        :param max_z: The amount to enlarge the maximum z-coordinate
        """
        self.min_x -= min_x
        self.min_y -= min_y
        self.min_z -= min_z
        self.max_x += max_x
        self.max_y += max_y
        self.max_z += max_z

    def enlarge_all(self, amount: float):
        """
        Enlarge the axis-aligned bounding box in all dimensions by a given amount in-place.

        :param amount: The amount to enlarge the bounding box
        """
        self.enlarge(amount, amount, amount,
                     amount, amount, amount)

    @classmethod
    def from_mesh(cls, mesh: trimesh.Trimesh) -> Self:
        """
        Create a bounding box from a trimesh object.
        :param mesh: The trimesh object.
        :return: The bounding box.
        """
        bounds = mesh.bounds
        return cls(bounds[0][0], bounds[0][1], bounds[0][2],
                   bounds[1][0], bounds[1][1], bounds[1][2])

    def get_points(self) -> List[Point3]:
        """
        Get the 8 corners of the bounding box as Point3 objects.

        :return: A list of Point3 objects representing the corners of the bounding box.
        """
        return [Point3(x, y, z)
                for x in (self.min_x, self.max_x)
                for y in (self.min_y, self.max_y)
                for z in (self.min_z, self.max_z)]

    @classmethod
    def from_min_max(cls, min_point: Point3, max_point: Point3) -> Self:
        """
        Set the axis-aligned bounding box from a minimum and maximum point.

        :param min_point: The minimum point
        :param max_point: The maximum point
        """
        return cls(*min_point.to_np()[:3], *max_point.to_np()[:3])



@dataclass
class BoundingBoxCollection:
    """
    Dataclass for storing a collection of bounding boxes.
    """

    bounding_boxes: List[BoundingBox] = field(default_factory=list)

    def __iter__(self) -> Iterator[BoundingBox]:
        return iter(self.bounding_boxes)

    @property
    def event(self) -> Event:
        """
        :return: The bounding boxes as a random event.
        """
        return Event(*[box.simple_event for box in self.bounding_boxes])

    def merge(self, other: BoundingBoxCollection) -> BoundingBoxCollection:
        """
        Merge another bounding box collection into this one.

        :param other: The other bounding box collection.
        :return: The merged bounding box collection.
        """
        return BoundingBoxCollection(self.bounding_boxes + other.bounding_boxes)

    def bloat(self, x_amount: float = 0., y_amount: float = 0, z_amount: float = 0) -> BoundingBoxCollection:
        """
        Enlarges all bounding boxes in the collection by a given amount in all dimensions.

        :param x_amount: The amount to adjust the x-coordinates
        :param y_amount: The amount to adjust the y-coordinates
        :param z_amount: The amount to adjust the z-coordinates

        :return: The enlarged bounding box collection
        """
        return BoundingBoxCollection([box.bloat(x_amount, y_amount, z_amount) for box in self.bounding_boxes])

    @classmethod
    def from_simple_event(cls, simple_event: SimpleEvent):
        """
        Create a list of bounding boxes from a simple random event.

        :param simple_event: The random event.
        :return: The list of bounding boxes.
        """
        result = []
        for x, y, z in itertools.product(simple_event[SpatialVariables.x.value].simple_sets,
                                         simple_event[SpatialVariables.y.value].simple_sets,
                                         simple_event[SpatialVariables.z.value].simple_sets):
            result.append(BoundingBox(x.lower, y.lower, z.lower, x.upper, y.upper, z.upper))
        return result

    @classmethod
    def from_event(cls, event: Event) -> Self:
        """
        Create a list of bounding boxes from a random event.

        :param event: The random event.
        :return: The list of bounding boxes.
        """
        return cls([box for simple_event in event.simple_sets for box in cls.from_simple_event(simple_event)])

    @classmethod
    def from_shapes(cls, shapes: List[Shape]) -> Self:
        """
        Create a bounding box collection from a list of shapes.

        :param shapes: The list of shapes.
        :return: The bounding box collection.
        """
        return cls([shape.as_bounding_box() for shape in shapes])



from typing import Dict, Callable, Union, Tuple, List

import numpy as np

from . import spatial_types as cas
from ..types import AnyMatrix4x4

Provider = Union[float, Callable[[], float]]

from threading import Lock


class SingletonMeta(type):
    """
    This is a thread-safe implementation of Singleton.
    Source: https://refactoring.guru/design-patterns/singleton/python/example#example-1
    """

    _instances = {}

    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__.py` argument do not affect
        the returned instance.
        """
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class SymbolManager(metaclass=SingletonMeta):
    """
    Singleton because I don't want multiple symbols with the same name to refer to different objects.
    TODO usually called registry
    Manages the association of symbolic variables with their providers and facilitates operations on them.

    The `SymbolManager` class is a tool for managing symbolic variables and their associated value providers.
    It allows the registration of various mathematical entities such as points, vectors, quaternions, and
    provides methods for resolving these symbols to their numeric values. The class also supports the
    evaluation of expressions involving these symbols.

    The purpose of this class is to abstract the management of symbolic variables and enable their seamless
    use in mathematical and symbolic computations.
    """
    symbol_to_provider: Dict[cas.Symbol, Callable[[], float]]
    """
    A dictionary mapping symbolic variables (`cas.Symbol`) to callable functions that provide numeric values for those symbols.
    """

    def __init__(self):
        self.symbol_to_provider = {}

    def register_symbol_provider(self, name: str, provider: Provider) -> cas.Symbol:
        """
        Creates a new symbol with the given name and associates it with a provider.
        """
        symbol = cas.Symbol(name)
        self.symbol_to_provider[symbol] = provider
        return symbol

    def register_point3(self, name: str, provider: Callable[[], Tuple[float, float, float]]) -> cas.Point3:
        """
        :param name: Used as prefix for the symbols x, y, and z.
        :param provider: A provider that returns a tuple of floats for x, y, and z.
        :return: A Point3 expression with symbols for x, y, and z.
        """
        sx = self.register_symbol_provider(f'{name}.x', lambda: provider()[0])
        sy = self.register_symbol_provider(f'{name}.y', lambda: provider()[1])
        sz = self.register_symbol_provider(f'{name}.z', lambda: provider()[2])
        p = cas.Point3(sx, sy, sz)
        return p

    def register_vector3(self, name: str, provider: Callable[[], Tuple[float, float, float]]) -> cas.Vector3:
        """
        :param name: Used as prefix for the symbols x, y, and z.
        :param provider: A provider that returns a tuple of floats for x, y, and z.
        :return: A Vector3 expression with symbols for x, y, and z.
        """
        sx = self.register_symbol_provider(f'{name}.x', lambda: provider()[0])
        sy = self.register_symbol_provider(f'{name}.y', lambda: provider()[1])
        sz = self.register_symbol_provider(f'{name}.z', lambda: provider()[2])
        v = cas.Vector3(sx, sy, sz)
        return v

    def register_quaternion(self, name: str, provider: Callable[[], Tuple[float, float, float, float]]) \
            -> cas.Quaternion:
        """
        :param name: Used as prefix for the symbols x, y, z, and w.
        :param provider: A provider that returns a tuple of floats for x, y, z, and w.
        :return: A Quaternion expression with symbols for x, y, z, and w.
        """
        sx = self.register_symbol_provider(f'{name}.x', lambda: provider()[0])
        sy = self.register_symbol_provider(f'{name}.y', lambda: provider()[1])
        sz = self.register_symbol_provider(f'{name}.z', lambda: provider()[2])
        sw = self.register_symbol_provider(f'{name}.w', lambda: provider()[3])
        q = cas.Quaternion(sx, sy, sz, sw)
        return q

    def register_transformation_matrix(self, name: str,
                                       provider: Callable[[], AnyMatrix4x4]) \
            -> cas.TransformationMatrix:
        """
        :param name: Used as prefix for the symbols.
        :param provider: A provider that returns a matrix.
                    A numpy 4x4 matrix is fine, but only a 3x4 matrix is required, as the last row is always [0,0,0,1].
        :return: A TransformationMatrix expression with symbols at all entries.
        """
        symbols = []
        for row in range(3):
            symbols.append([])
            for col in range(4):
                symbols[row].append(self.register_symbol_provider(f'{name}[{row},{col}]',
                                                                  lambda r=row, c=col: provider()[r][c]))
        symbols.append([0,0,0,1])
        root_T_tip = cas.TransformationMatrix(symbols)
        return root_T_tip

    def resolve_symbols(self, symbols: Union[List[cas.Symbol], List[List[cas.Symbol]]]) \
            -> Union[np.ndarray, List[np.ndarray]]:
        """
        Given a list of symbols or a list of lists of symbols, returns a list of numeric values for each symbol computed using their providers.
        Intended to be used in combination with `CompiledFunction.symbol_parameters` and `CompiledFunction.fast_call`.
        """
        try:
            if len(symbols) == 0:
                return np.array([])
            if isinstance(symbols[0], list):
                return [np.array([self.symbol_to_provider[s]() for s in param], dtype=float) for param in symbols]
            else:
                return np.array([self.symbol_to_provider[s]() for s in symbols], dtype=float)
        except Exception as e:
            # Flatten symbols for error checking
            flattened_symbols = []
            if len(symbols) > 0 and isinstance(symbols[0], list):
                # symbols is a list of lists
                for sublist in symbols:
                    flattened_symbols.extend(sublist)
            else:
                # symbols is already a flat list
                flattened_symbols = symbols

            for s in flattened_symbols:
                try:
                    np.array([self.symbol_to_provider[s]()])
                except Exception as e2:
                    raise KeyError(f'Cannot resolve {s} ({e2.__class__.__name__}: {str(e2)})')
            raise e

    def evaluate_expr(self, expr: cas.Expression):
        """
        Compiles and evaluates an arbitrary symbolic expression, using the current symbol providers.
        """
        if isinstance(expr, (int, float)):
            return expr
        f = expr.compile()
        if len(f.symbol_parameters) == 0:
            return expr.to_np()
        result = f.fast_call(*self.resolve_symbols(f.symbol_parameters))
        if len(result) == 1:
            return result[0]
        else:
            return result


symbol_manager = SymbolManager()



from __future__ import annotations

import atexit
import logging
import threading
import time
from dataclasses import dataclass

# from roslaunch.core import get_ros_package_path as get_ros_package_paths

from ..geometry import Shape, Mesh, Sphere, Cylinder, Box, Color
from ..spatial_types import Vector3, Quaternion
from ..utils import IDGenerator
from ..world import World

try:
    import rospy
    from rospy.rostime import Time as ROSTime
    from visualization_msgs.msg import MarkerArray, Marker
    from geometry_msgs.msg import (Vector3 as ROSVector3, Quaternion as ROSQuaternion, Pose as ROSPose,
                                   PoseStamped as ROSPoseStamped, Header as ROSHeader, Color as ROSColor, )
except ImportError:
    logging.warn("Importing ros adapter without rospy available.")


def get_ros_package_path(package_name: str) -> str:
    """
    Placeholder for ros integration
    """
    return [p for p in get_ros_package_paths().split(':') if p.endswith(package_name)][0]


id_generator = IDGenerator()

#
# def vector3_to_ros_message(obj: Vector3) -> ROSVector3:
#     return ROSVector3(x=obj.x, y=obj.y, z=obj.z)
#
#
# def quaternion_to_ros_message(obj: Quaternion) -> ROSQuaternion:
#     return ROSQuaternion(x=obj.x, y=obj.y, z=obj.z, w=obj.w)
#
#
# def pose_to_ros_message(obj: Pose) -> ROSPose:
#     return ROSPose(position=vector3_to_ros_message(obj.position),
#                    orientation=quaternion_to_ros_message(obj.orientation))
#
#
# def header_to_ros_message(obj: Header) -> ROSHeader:
#     stamp = ROSTime.from_sec(obj.timestamp.timestamp())
#     return ROSHeader(frame_id=obj.frame_id, stamp=stamp, seq=obj.sequence)
#
#
# def pose_stamped_to_ros_message(obj: PoseStamped) -> PoseStamped:
#     return ROSPoseStamped(pose=pose_to_ros_message(obj.pose), header=header_to_ros_message(obj.header))
#
#
# def color_to_ros_message(obj: Color):
#     return ROSColor(rgba=vector3_to_ros_message(obj.rgba), )


def shape_to_ros_message(obj: Shape) -> Marker:
    marker = Marker()
    marker.header = header_to_ros_message(obj.origin.header)
    marker.action = Marker.ADD
    marker.pose = pose_to_ros_message(obj.origin.pose)
    marker.scale = vector3_to_ros_message(Vector3(1., 1., 1.))
    marker.color = color_to_ros_message(Color())
    marker.lifetime = rospy.Duration.from_sec(10)
    marker.ns = "shapes"
    marker.id = id_generator(id(obj))

    return adapter_methods[type(obj)](obj, marker)


def mesh_to_ros_message(obj: Mesh, marker: Marker) -> Marker:
    marker.type = Marker.MESH_RESOURCE
    marker.mesh_resource = "file://" + obj.filename
    marker.mesh_use_embedded_materials = True
    marker.scale = vector3_to_ros_message(obj.scale)
    return marker


def sphere_to_ros_message(obj: Sphere, marker: Marker) -> Marker:
    marker.color = color_to_ros_message(obj.color)
    marker.type = Marker.SPHERE
    marker.scale = vector3_to_ros_message(Vector3(obj.radius * 2, obj.radius * 2, obj.radius * 2))
    return marker


def cylinder_to_ros_message(obj: Cylinder, marker: Marker) -> Marker:
    marker.color = color_to_ros_message(obj.color)
    marker.type = Marker.CYLINDER
    marker.scale = vector3_to_ros_message(Vector3(obj.radius * 2, obj.radius * 2, obj.length))
    return marker


def box_to_ros_message(obj: Box, marker: Marker) -> Marker:
    marker.color = color_to_ros_message(obj.color)
    marker.type = Marker.CUBE
    marker.scale = vector3_to_ros_message(Vector3(obj.length, obj.width, obj.height))
    return marker


adapter_methods = {
    Mesh: mesh_to_ros_message,
    Sphere: sphere_to_ros_message,
    Cylinder: cylinder_to_ros_message,
    Box: box_to_ros_message,
}


@dataclass
class WorldPublisher:
    """
    Publishes the visuals of every link in the world to a ros topic.
    The Publisher creates an Array of Visualization marker with a Marker for each link of each Object in the
    World. This Array is published with a rate of `interval`.
    The publisher runs in its own thread.
    """

    world: World
    """
    The world to read the links from.
    """

    topic_name: str = "/world/viz_marker"
    """
    The name of the topic to publish the visuals to.
    """

    interval: float = 0.1
    """
    The interval in seconds at which to publish the visuals.
    """

    def __post_init__(self):
        self.pub = rospy.Publisher(self.topic_name, MarkerArray, queue_size=10)
        self.thread = threading.Thread(target=self._publish, name="WorldPublisher")
        self.kill_event = threading.Event()
        self.thread.start()
        atexit.register(self._stop_publishing)

    def _publish(self) -> None:
        """
        Constantly publishes the Marker Array. To the given topic name at a fixed rate.
        """
        while not self.kill_event.is_set():
            marker_array = self._make_marker_array()
            self.pub.publish(marker_array)
            time.sleep(self.interval)

    def _make_marker_array(self) -> MarkerArray:
        """
        Creates the Marker Array to be published. There is one Marker for link for each object in the Array, each Object
        creates a name space in the visualization Marker. The type of Visualization Marker is decided by the collision
        tag of the URDF.

        :return: An Array of Visualization Marker
        """
        marker_array = MarkerArray()
        for body in self.world.bodies:
            for shape in body.visual:
                marker = shape.ros_message()
                marker_array.markers.append(marker)
        return marker_array

    def _stop_publishing(self) -> None:
        """
        Stops the publishing of the Visualization Marker update by setting the kill event and collecting the thread.
        """
        self.kill_event.set()
        self.thread.join()


from dataclasses import dataclass, field

import numpy as np
from probabilistic_model.probabilistic_circuit.rx.helper import uniform_measure_of_event
from typing_extensions import List

from semantic_world.geometry import BoundingBox, BoundingBoxCollection
from semantic_world.prefixed_name import PrefixedName
from semantic_world.spatial_types import Point3
from semantic_world.variables import SpatialVariables
from semantic_world.world import View, Body


@dataclass(unsafe_hash=True)
class Handle(View):
    body: Body

    def __post_init__(self):
        self.name = PrefixedName(str(self.body.name), self.__class__.__name__)


@dataclass(unsafe_hash=True)
class Container(View):
    body: Body

    def __post_init__(self):
        self.name = PrefixedName(str(self.body.name), self.__class__.__name__)


@dataclass(unsafe_hash=True)
class Door(View):  # Door has a Footprint
    """
    Door in a body that has a Handle and can open towards or away from the user.
    """
    handle: Handle
    body: Body

    def __post_init__(self):
        self.name = PrefixedName(str(self.body.name), self.__class__.__name__)

@dataclass(unsafe_hash=True)
class Fridge(View):
    body: Body
    door: Door

    def __post_init__(self):
        self.name = PrefixedName(str(self.body.name), self.__class__.__name__)

@dataclass(unsafe_hash=True)
class Table(View):
    """
    A view that represents a table.
    """

    top: Body
    """
    The body that represents the table's top surface.
    """

    def points_on_table(self, amount: int = 100) -> List[Point3]:
        """
        Get points that are on the table.

        :amount: The number of points to return.
        :returns: A list of points that are on the table.
        """
        area_of_table = BoundingBoxCollection.from_shapes(self.top.collision)
        event = area_of_table.event
        p = uniform_measure_of_event(event)
        p = p.marginal(SpatialVariables.xy)
        samples = p.sample(amount)
        z_coordinate = np.full((amount, 1), max([b.max_z for b in area_of_table]) + 0.01)
        samples = np.concatenate((samples, z_coordinate), axis=1)
        return [Point3(*s, reference_frame=self.top) for s in samples]

    def __post_init__(self):
        self.name = self.top.name

################################


@dataclass(unsafe_hash=True)
class Components(View):
    ...


@dataclass(unsafe_hash=True)
class Furniture(View):
    ...


#################### subclasses von Components


@dataclass(unsafe_hash=True)
class Door(Components):
    body: Body
    handle: Handle

    def __post_init__(self):
        self.name = PrefixedName(str(self.body.name), self.__class__.__name__)


@dataclass(unsafe_hash=True)
class Drawer(Components):
    container: Container
    handle: Handle

    def __post_init__(self):
        self.name = self.container.name


############################### subclasses to Furniture
@dataclass
class Cupboard(Furniture):
    ...


############################### subclasses to Cupboard
@dataclass(unsafe_hash=True)
class Cabinet(Cupboard):
    container: Container
    drawers: list[Drawer] = field(default_factory=list, hash=False)

    def __post_init__(self):
        self.name = self.container.name


@dataclass
class Wardrobe(Cupboard):
    doors: List[Door] = field(default_factory=list)



from __future__ import absolute_import
from __future__ import annotations

import logging
from copy import deepcopy
from dataclasses import dataclass, field
from enum import IntEnum
from functools import wraps, lru_cache
from typing import Dict, Tuple, OrderedDict, Union, Optional

import matplotlib.pyplot as plt
import numpy as np
import rustworkx as rx
import rustworkx.visit
import rustworkx.visualization
from typing_extensions import List, Type

from .connections import HasUpdateState, Has1DOFState, Connection6DoF
from .degree_of_freedom import DegreeOfFreedom
from .exceptions import DuplicateViewError, AddingAnExistingViewError, ViewNotFoundError
from .ik_solver import InverseKinematicsSolver
from .prefixed_name import PrefixedName
from .spatial_types import spatial_types as cas
from .spatial_types.derivatives import Derivatives, DerivativeMap
from .spatial_types.math import inverse_frame
from .types import NpMatrix4x4
from .utils import IDGenerator, copy_lru_cache
from .world_entity import Body, Connection, View
from .world_state import WorldState

logger = logging.getLogger(__name__)

id_generator = IDGenerator()


class PlotAlignment(IntEnum):
    HORIZONTAL = 0
    VERTICAL = 1


class ForwardKinematicsVisitor(rustworkx.visit.DFSVisitor):
    """
    Visitor class for collection various forward kinematics expressions in a world model.

    This class is designed to traverse a world, compute the forward kinematics transformations in batches for different
    use cases.
    1. Efficient computation of forward kinematics between any bodies in the world.
    2. Efficient computation of forward kinematics for all bodies with collisions for updating collision checkers.
    3. Efficient computation of forward kinematics as position and quaternion, useful for ROS tf.
    """

    compiled_collision_fks: cas.CompiledFunction
    compiled_all_fks: cas.CompiledFunction

    forward_kinematics_for_all_bodies: np.ndarray
    """
    A 2D array containing the stacked forward kinematics expressions for all bodies in the world.
    Dimensions are ((number of bodies) * 4) x 4.
    They are computed in batch for efficiency.
    """
    body_name_to_forward_kinematics_idx: Dict[PrefixedName, int]
    """
    Given a body name, returns the index of the first row in `forward_kinematics_for_all_bodies` that corresponds to that body.
    """

    def __init__(self, world: World):
        self.world = world
        self.child_body_to_fk_expr: Dict[PrefixedName, cas.TransformationMatrix] = {
            self.world.root.name: cas.TransformationMatrix()}
        self.tf: Dict[Tuple[PrefixedName, PrefixedName], cas.Expression] = OrderedDict()

    def connection_call(self, edge: Tuple[int, int, Connection]):
        """
        Gathers forward kinematics expressions for a connection.
        """
        connection = edge[2]
        map_T_parent = self.child_body_to_fk_expr[connection.parent.name]
        self.child_body_to_fk_expr[connection.child.name] = map_T_parent.dot(connection.origin_expression)
        self.tf[(connection.parent.name, connection.child.name)] = connection.origin_as_position_quaternion()

    tree_edge = connection_call

    def compile_forward_kinematics(self) -> None:
        """
        Compiles forward kinematics expressions for fast evaluation.
        """
        all_fks = cas.vstack([self.child_body_to_fk_expr[body.name] for body in self.world.bodies])
        tf = cas.vstack([pose for pose in self.tf.values()])
        collision_fks = []
        for body in self.world.bodies_with_collisions:
            if body == self.world.root:
                continue
            collision_fks.append(self.child_body_to_fk_expr[body.name])
        collision_fks = cas.vstack(collision_fks)
        params = [v.symbols.position for v in self.world.degrees_of_freedom]
        self.compiled_all_fks = all_fks.compile(parameters=params)
        self.compiled_collision_fks = collision_fks.compile(parameters=params)
        self.compiled_tf = tf.compile(parameters=params)
        self.idx_start = {body.name: i * 4 for i, body in enumerate(self.world.bodies)}

    def recompute(self) -> None:
        """
        Clears cache and recomputes all forward kinematics. Should be called after a state update.
        """
        self.compute_forward_kinematics_np.cache_clear()
        self.subs = self.world.state.positions
        self.forward_kinematics_for_all_bodies = self.compiled_all_fks.fast_call(self.subs)

    def compute_tf(self) -> np.ndarray:
        """
        Computes a (number of bodies) x 7 matrix of forward kinematics in position/quaternion format.
        The rows are ordered by body name.
        The first 3 entries are position values, the last 4 entires are quaternion values in x, y, z, w order.

        This is not updated in 'recompute', because this functionality is only used with ROS.
        :return: A large matrix with all forward kinematics.
        """
        return self.compiled_tf.fast_call(self.subs)

    @lru_cache(maxsize=None)
    def compute_forward_kinematics_np(self, root: Body, tip: Body) -> NpMatrix4x4:
        """
        Computes the forward kinematics from the root body to the tip body, root_T_tip.

        This method computes the transformation matrix representing the pose of the
        tip body relative to the root body, expressed as a numpy ndarray.

        :param root: Root body for which the kinematics are computed.
        :param tip: Tip body to which the kinematics are computed.
        :return: Transformation matrix representing the relative pose of the tip body with respect to the root body.
        """
        root = root.name
        tip = tip.name
        root_is_world = root == self.world.root.name
        tip_is_world = tip == self.world.root.name

        if not tip_is_world:
            i = self.idx_start[tip]
            map_T_tip = self.forward_kinematics_for_all_bodies[i:i + 4]
            if root_is_world:
                return map_T_tip

        if not root_is_world:
            i = self.idx_start[root]
            map_T_root = self.forward_kinematics_for_all_bodies[i:i + 4]
            root_T_map = inverse_frame(map_T_root)
            if tip_is_world:
                return root_T_map

        if tip_is_world and root_is_world:
            return np.eye(4)

        return root_T_map @ map_T_tip


class ResetStateContextManager:
    """
    A context manager for resetting the state of a given `World` instance.

    This class is designed to allow operations to be performed on a `World`
    object, ensuring that its state can be safely returned to its previous
    condition upon leaving the context. If no exceptions occur within the
    context, the original state of the `World` instance is restored, and the
    state change is notified.
    """

    def __init__(self, world: World):
        self.world = world

    def __enter__(self) -> None:
        self.state = deepcopy(self.world.state)

    def __exit__(self, exc_type: Optional[type], exc_val: Optional[Exception], exc_tb: Optional[type]) -> None:
        if exc_type is None:
            self.world.state = self.state
            self.world.notify_state_change()


class WorldModelUpdateContextManager:
    """
    Context manager for updating the state of a given `World` instance.
    This class manages that updates to the world within the context of this class only trigger recomputations after all
    desired updates have been performed.
    """
    first: bool = True

    def __init__(self, world: World):
        self.world = world

    def __enter__(self):
        if self.world.world_is_being_modified:
            self.first = False
        self.world.world_is_being_modified = True
        return self.world

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.first:
            self.world.world_is_being_modified = False
            if exc_type is None:
                self.world._notify_model_change()


def modifies_world(func):
    """
    Decorator that marks a method as a modification to the state or model of a world.
    """

    @wraps(func)
    def wrapper(self: World, *args, **kwargs):
        with self.modify_world() as context_manager:
            result = func(self, *args, **kwargs)
            return result

    return wrapper


@dataclass
class World:
    """
    A class representing the world.
    The world manages a set of bodies and connections represented as a tree-like graph.
    The nodes represent bodies in the world, and the edges represent joins between them.
    """

    kinematic_structure: rx.PyDAG[Body] = field(default_factory=lambda: rx.PyDAG(multigraph=False), kw_only=True,
                                                repr=False)
    """
    The kinematic structure of the world.
    The kinematic structure is a tree shaped directed graph where the nodes represent bodies in the world,
    and the edges represent connections between them.
    """

    views: List[View] = field(default_factory=list, repr=False)
    """
    All views the world is aware of.
    """

    degrees_of_freedom: List[DegreeOfFreedom] = field(default_factory=list)

    state: WorldState = field(default_factory=WorldState)
    """
    2d array where rows are derivatives and columns are dof values for that derivative.
    """

    _model_version: int = 0
    """
    The version of the model. This increases whenever a change to the kinematic model is made. Mostly triggered
    by adding/removing bodies and connections.
    """

    _state_version: int = 0
    """
    The version of the state. This increases whenever a change to the state of the kinematic model is made. 
    Mostly triggered by updating connection values.
    """

    world_is_being_modified: bool = False
    """
    Is set to True, when a function with @modifies_world is called or world.modify_world context is used.
    """

    name: Optional[str] = None
    """
    Name of the world. May act as default namespace for all bodies and views in the world which do not have a prefix.
    """

    @property
    def root(self) -> Body:
        """
        The root of the world is the unique node with in-degree 0.

        :return: The root of the world.
        """
        possible_roots = [node for node in self.bodies if self.kinematic_structure.in_degree(node.index) == 0]
        if len(possible_roots) == 1:
            return possible_roots[0]
        elif len(possible_roots) > 1:
            raise ValueError(f"More than one root found. Possible roots are {possible_roots}")
        else:
            raise ValueError(f"No root found.")

    def __hash__(self):
        return hash(id(self))

    def validate(self) -> bool:
        """
        Validate the world.

        The world must be a tree.
        :return: True if the world is valid, raises an AssertionError otherwise.
        """
        assert len(self.bodies) == (len(self.connections) + 1)
        assert rx.is_weakly_connected(self.kinematic_structure)
        return True

    @modifies_world
    def create_degree_of_freedom(self, name: PrefixedName, lower_limits: Optional[DerivativeMap[float]] = None,
                                 upper_limits: Optional[DerivativeMap[float]] = None) -> DegreeOfFreedom:
        """
        Create a degree of freedom in the world and return it.
        For dependent kinematics, DoFs must be created with this method and passed to the connection's conctructor.
        :param name: Name of the DoF.
        :param lower_limits: If the DoF is actively controlled, it must have at least velocity limits.
        :param upper_limits: If the DoF is actively controlled, it must have at least velocity limits.
        :return: The already registered DoF.
        """
        dof = DegreeOfFreedom(name=name, lower_limits=lower_limits, upper_limits=upper_limits, _world=self)
        initial_position = 0
        lower_limit = dof.lower_limits.position
        if lower_limit is not None:
            initial_position = max(lower_limit, initial_position)
        upper_limit = dof.upper_limits.position
        if upper_limit is not None:
            initial_position = min(upper_limit, initial_position)
        self.state[name].position = initial_position
        assert [dof for dof in self.degrees_of_freedom if dof.name == name].count(dof) == 0
        self.degrees_of_freedom.append(dof)
        return dof

    def modify_world(self) -> WorldModelUpdateContextManager:
        return WorldModelUpdateContextManager(self)

    def reset_state_context(self) -> ResetStateContextManager:
        return ResetStateContextManager(self)

    def reset_cache(self) -> None:
        # super().reset_cache()
        # self.get_directly_controlled_child_links_with_collisions.cache_clear()
        # self.get_directly_controlled_child_links_with_collisions.cache_clear()
        # self.compute_chain_reduced_to_controlled_joints.cache_clear()
        # self.get_movable_parent_joint.cache_clear()
        # self.get_controlled_parent_joint_of_link.cache_clear()
        # self.get_controlled_parent_joint_of_joint.cache_clear()
        self.compute_split_chain_of_bodies.cache_clear()
        self.compute_split_chain_of_connections.cache_clear()
        # self.are_linked.cache_clear()
        # self.compose_fk_expression.cache_clear()
        self.compute_chain_of_bodies.cache_clear()
        self.compute_chain_of_connections.cache_clear()
        # self.is_link_controlled.cache_clear()
        for dof in self.degrees_of_freedom:
            dof.reset_cache()

    def notify_state_change(self) -> None:
        """
        If you have changed the state of the world, call this function to trigger necessary events and increase
        the state version.
        """
        # self.compute_fk.cache_clear()
        # self.compute_fk_with_collision_offset_np.cache_clear()
        self._recompute_forward_kinematics()
        self._state_version += 1

    def _notify_model_change(self) -> None:
        """
        Call this function if you have changed the model of the world to trigger necessary events and increase
        the model version number.
        """
        if not self.world_is_being_modified:
            # self._fix_tree_structure()
            self.reset_cache()
            self.compile_forward_kinematics_expressions()
            # self._cleanup_unused_dofs()
            self.notify_state_change()
            self._model_version += 1
            self.validate()

    @property
    def bodies(self) -> List[Body]:
        """
        :return: A list of all bodies in the world.
        """
        return list(self.kinematic_structure.nodes())

    @property
    def bodies_with_collisions(self) -> List[Body]:
        """
        :return: A list of all bodies in the world that have collisions.
        """
        return [body for body in self.bodies if body.has_collision()]

    @property
    def connections(self) -> List[Connection]:
        """
        :return: A list of all connections in the world.
        """
        return list(self.kinematic_structure.edges())

    @modifies_world
    def add_body(self, body: Body) -> None:
        """
        Add a body to the world.

        :param body: The body to add.
        """
        if body._world is self and body.index is not None:
            return
        elif body._world is not None and body._world is not self:
            raise NotImplementedError("Cannot add a body that already belongs to another world.")

        body.index = self.kinematic_structure.add_node(body)

        # write self as the bodys world
        body._world = self

    @modifies_world
    def add_connection(self, connection: Connection) -> None:
        """
        Add a connection and the bodies it connects to the world.

        :param connection: The connection to add.
        """
        self.add_body(connection.parent)
        self.add_body(connection.child)
        connection._world = self
        self.kinematic_structure.add_edge(connection.parent.index, connection.child.index, connection)

    def add_view(self, view: View, exists_ok: bool = False) -> None:
        """
        Adds a view to the current list of views if it doesn't already exist. Ensures
        that the `view` is associated with the current instance and maintains the
        integrity of unique view names.

        :param view: The view instance to be added. Its name must be unique within
            the current context.
        :param exists_ok: Whether to raise an error or not when a view already exists.

        :raises AddingAnExistingViewError: If exists_ok is False and a view with the same name and type already exists.
        """
        try:
            self.get_view_by_name(view.name)
            if not exists_ok:
                raise AddingAnExistingViewError(view)
        except ViewNotFoundError:
            view._world = self
            self.views.append(view)

    def get_view_by_name(self, name: Union[str, PrefixedName]) -> Optional[View]:
        """
        Retrieves a View from the list of view based on its name.
        If the input is of type `PrefixedName`, it checks whether the prefix is specified and looks for an
        exact match. Otherwise, it matches based on the name's string representation.
        If more than one body with the same name is found, an assertion error is raised.
        If no matching body is found, a `ValueError` is raised.

        :param name: The name of the view to search for. Can be a string or a `PrefixedName` object.
        :return: The `View` object that matches the given name.
        """
        if isinstance(name, PrefixedName):
            if name.prefix is not None:
                matches = [view for view in self.views if view.name == name]
            else:
                matches = [view for view in self.views if view.name.name == name.name]
        else:
            matches = [view for view in self.views if view.name.name == name]
        if len(matches) > 1:
            raise DuplicateViewError(matches)
        if matches:
            return matches[0]
        raise ViewNotFoundError(name)

    @modifies_world
    def remove_body(self, body: Body) -> None:
        if body._world is self and body.index is not None:
            self.kinematic_structure.remove_node(body.index)
            body._world = None
            body.index = None
        else:
            logger.debug("Trying to remove a body that is not part of this world.")

    @modifies_world
    def merge_world(self, other: World, root_connection: Connection = None) -> None:
        """
        Merge a world into the existing one by merging degrees of freedom, states, connections, and bodies.
        This removes all bodies and connections from `other`.

        :param other: The world to be added.
        :param root_connection: If provided, this connection will be used to connect the two worlds. Otherwise, a new Connection6DoF will be created
        :return: None
        """

        self_root = self.root
        other_root = other.root
        for dof in other.degrees_of_freedom:
            self.state[dof.name].position = other.state[dof.name].position
            self.state[dof.name].velocity = other.state[dof.name].velocity
            self.state[dof.name].acceleration = other.state[dof.name].acceleration
            self.state[dof.name].jerk = other.state[dof.name].jerk
            dof._world = self
        self.degrees_of_freedom.extend(other.degrees_of_freedom)

        # do not trigger computations in other
        other.world_is_being_modified = True
        for connection in other.connections:
            other.remove_body(connection.parent)
            other.remove_body(connection.child)
            self.add_connection(connection)
        other.world_is_being_modified = False

        connection = root_connection or Connection6DoF(parent=self_root, child=other_root, _world=self)
        self.add_connection(connection)

    def merge_world_at_pose(self, other: World, pose: cas.TransformationMatrix) -> None:
        """
        Merge another world into the existing one, creates a 6DoF connection between the root of this world and the root
        of the other world.
        :param other: The world to be added.
        :param pose: world_root_T_other_root, the pose of the other world's root with respect to the current world's root
        """
        root_connection = Connection6DoF(parent=self.root, child=other.root, _world=self)
        root_connection.origin = pose
        self.merge_world(other, root_connection)
        self.add_connection(root_connection)

    def __str__(self):
        return f"{self.__class__.__name__} with {len(self.bodies)} bodies."

    def get_connection(self, parent: Body, child: Body) -> Connection:
        return self.kinematic_structure.get_edge_data(parent.index, child.index)

    def get_body_by_name(self, name: Union[str, PrefixedName]) -> Body:
        """
        Retrieves a body from the list of bodies based on its name.
        If the input is of type `PrefixedName`, it checks whether the prefix is specified and looks for an
        exact match. Otherwise, it matches based on the name's string representation.
        If more than one body with the same name is found, an assertion error is raised.
        If no matching body is found, a `ValueError` is raised.

        :param name: The name of the body to search for. Can be a string or a `PrefixedName` object.
        :return: The `Body` object that matches the given name.
        :raises ValueError: If multiple or no bodies with the specified name are found.
        """
        if isinstance(name, PrefixedName):
            if name.prefix is not None:
                matches = [body for body in self.bodies if body.name == name]
            else:
                matches = [body for body in self.bodies if body.name.name == name.name]
        else:
            matches = [body for body in self.bodies if body.name.name == name]
        if len(matches) > 1:
            raise ValueError(f'Multiple bodies with name {name} found')
        if matches:
            return matches[0]
        raise KeyError(f'Body with name {name} not found')

    def get_degree_of_freedom_by_name(self, name: Union[str, PrefixedName]) -> DegreeOfFreedom:
        """
        Retrieves a DegreeOfFreedom from the list of DegreeOfFreedom based on its name.
        If the input is of type `PrefixedName`, it checks whether the prefix is specified and looks for an
        exact match. Otherwise, it matches based on the name's string representation.
        If more than one body with the same name is found, an assertion error is raised.
        If no matching body is found, a `ValueError` is raised.

        :param name: The name of the DegreeOfFreedom to search for. Can be a string or a `PrefixedName` object.
        :return: The `DegreeOfFreedom` object that matches the given name.
        :raises ValueError: If multiple or no DegreeOfFreedom with the specified name are found.
        """
        if isinstance(name, PrefixedName):
            if name.prefix is not None:
                matches = [dof for dof in self.degrees_of_freedom if dof.name == name]
            else:
                matches = [dof for dof in self.degrees_of_freedom if dof.name.name == name.name]
        else:
            matches = [dof for dof in self.degrees_of_freedom if dof.name.name == name]
        if len(matches) > 1:
            raise ValueError(f'Multiple DegreeOfFreedom with name {name} found')
        if matches:
            return matches[0]
        raise KeyError(f'DegreeOfFreedom with name {name} not found')

    def get_connection_by_name(self, name: Union[str, PrefixedName]) -> Connection:
        """
        Retrieve a connection by its name.
        This method accepts either a string or a `PrefixedName` instance.
        It searches through the list of connections and returns the one
        that matches the given name. If the `PrefixedName` contains a prefix,
        the method ensures the name, including the prefix, matches an existing
        connection. Otherwise, it only considers the unprefixed name. If more than
        one connection matches the specified name, or if no connection is found,
        an exception is raised.

        :param name: The name of the connection to retrieve. Can be a string or
            a `PrefixedName` instance. If a prefix is included in `PrefixedName`,
            it will be used for matching.
        :return: The connection that matches the specified name.
        :raises ValueError: If multiple connections with the given name are found
            or if no connection with the given name exists.
        """
        if isinstance(name, PrefixedName):
            if name.prefix is not None:
                matches = [conn for conn in self.connections if conn.name == name]
            else:
                matches = [conn for conn in self.connections if conn.name.name == name.name]
        else:
            matches = [conn for conn in self.connections if conn.name.name == name]
        if len(matches) > 1:
            raise ValueError(f'Multiple connections with name {name} found')
        if matches:
            return matches[0]
        raise KeyError(f'Connection with name {name} not found')

    @lru_cache(maxsize=None)
    def compute_child_bodies(self, body: Body) -> List[Body]:
        """
        Computes the child bodies of a given body in the world.
        :param body: The body for which to compute child bodies.
        :return: A list of child bodies.
        """
        return list(self.kinematic_structure.successors(body.index))

    def compute_child_bodies_recursive(self, body: Body) -> List[Body]:
        """
        Computes all child bodies of a given body in the world recursively.
        :param body: The body for which to compute child bodies.
        :return: A list of all child bodies.
        """
        children = self.compute_child_bodies(body)
        for child in children:
            children.extend(self.compute_child_bodies_recursive(child))
        return children

    @lru_cache(maxsize=None)
    def compute_parent_body(self, body: Body) -> Body:
        """
        Computes the parent body of a given body in the world.
        :param body: The body for which to compute the parent body.
        :return: The parent body of the given body.
        """
        return next(iter(self.kinematic_structure.predecessors(body.index)))

    @lru_cache(maxsize=None)
    def compute_parent_connection(self, body: Body) -> Connection:
        """
        Computes the parent connection of a given body in the world.
        :param body: The body for which to compute the parent connection.
        :return: The parent connection of the given body.
        """
        return self.kinematic_structure.get_edge_data(self.compute_parent_body(body).index, body.index)

    @lru_cache(maxsize=None)
    def compute_chain_of_bodies(self, root: Body, tip: Body) -> List[Body]:
        if root == tip:
            return [root]
        shortest_paths = rx.all_shortest_paths(self.kinematic_structure, root.index, tip.index, as_undirected=False)

        if len(shortest_paths) == 0:
            raise rx.NoPathFound(f'No path found from {root} to {tip}')

        return [self.kinematic_structure[index] for index in shortest_paths[0]]

    @lru_cache(maxsize=None)
    def compute_chain_of_connections(self, root: Body, tip: Body) -> List[Connection]:
        body_chain = self.compute_chain_of_bodies(root, tip)
        return [self.get_connection(body_chain[i], body_chain[i + 1]) for i in range(len(body_chain) - 1)]

    @lru_cache(maxsize=None)
    def compute_split_chain_of_bodies(self, root: Body, tip: Body) -> Tuple[List[Body], List[Body], List[Body]]:
        """
        Computes the chain between root and tip. Can handle chains that start and end anywhere in the tree.
        :param root: The root body to start the chain from
        :param tip: The tip body to end the chain at
        :return: tuple containing
                    1. chain from root to the common ancestor (excluding common ancestor)
                    2. list containing just the common ancestor
                    3. chain from common ancestor to tip (excluding common ancestor)
        """
        if root == tip:
            return [], [root], []
        root_chain = self.compute_chain_of_bodies(self.root, root)
        tip_chain = self.compute_chain_of_bodies(self.root, tip)
        i = 0
        for i in range(min(len(root_chain), len(tip_chain))):
            if root_chain[i] != tip_chain[i]:
                break
        else:
            i += 1
        common_ancestor = tip_chain[i - 1]
        root_chain = self.compute_chain_of_bodies(common_ancestor, root)
        root_chain = root_chain[1:]
        root_chain = root_chain[::-1]
        tip_chain = self.compute_chain_of_bodies(common_ancestor, tip)
        tip_chain = tip_chain[1:]
        return root_chain, [common_ancestor], tip_chain

    @lru_cache(maxsize=None)
    def compute_split_chain_of_connections(self, root: Body, tip: Body) -> Tuple[List[Connection], List[Connection]]:
        """
        Computes split chains of connections between 'root' and 'tip' bodies. Returns tuple of two Connection lists:
        (root->common ancestor, tip->common ancestor). Returns empty lists if root==tip.

        :param root: The starting `Body` object for the chain of connections.
        :param tip: The ending `Body` object for the chain of connections.
        :return: A tuple of two lists: the first list contains `Connection` objects from the `root` to
            the common ancestor, and the second list contains `Connection` objects from the `tip` to the
            common ancestor.
        """
        if root == tip:
            return [], []
        root_chain, common_ancestor, tip_chain = self.compute_split_chain_of_bodies(root, tip)
        root_chain.append(common_ancestor[0])
        tip_chain.insert(0, common_ancestor[0])
        root_connections = []
        for i in range(len(root_chain) - 1):
            root_connections.append(self.get_connection(root_chain[i + 1], root_chain[i]))
        tip_connections = []
        for i in range(len(tip_chain) - 1):
            tip_connections.append(self.get_connection(tip_chain[i], tip_chain[i + 1]))
        return root_connections, tip_connections

    @property
    def layers(self) -> List[List[Body]]:
        return rx.layers(self.kinematic_structure, [self.root.index], index_output=False)

    def bfs_layout(self, scale: float = 1., align: PlotAlignment = PlotAlignment.VERTICAL) -> Dict[int, np.array]:
        """
        Generate a bfs layout for this circuit.

        :return: A dict mapping the node indices to 2d coordinates.
        """
        layers = self.layers

        pos = None
        nodes = []
        width = len(layers)
        for i, layer in enumerate(layers):
            height = len(layer)
            xs = np.repeat(i, height)
            ys = np.arange(0, height, dtype=float)
            offset = ((width - 1) / 2, (height - 1) / 2)
            layer_pos = np.column_stack([xs, ys]) - offset
            if pos is None:
                pos = layer_pos
            else:
                pos = np.concatenate([pos, layer_pos])
            nodes.extend(layer)

        # Find max length over all dimensions
        pos -= pos.mean(axis=0)
        lim = np.abs(pos).max()  # max coordinate for all axes
        # rescale to (-scale, scale) in all directions, preserves aspect
        if lim > 0:
            pos *= scale / lim

        if align == PlotAlignment.HORIZONTAL:
            pos = pos[:, ::-1]  # swap x and y coords

        pos = dict(zip([node.index for node in nodes], pos))
        return pos

    def plot_kinematic_structure(self, scale: float = 1., align: PlotAlignment = PlotAlignment.VERTICAL) -> None:
        """
        Plots the kinematic structure of the world.
        The plot shows bodies as nodes and connections as edges in a directed graph.
        """
        # Create a new figure
        plt.figure(figsize=(12, 8))

        pos = self.bfs_layout(scale=scale, align=align)

        rustworkx.visualization.mpl_draw(self.kinematic_structure, pos=pos, labels=lambda body: str(body.name),
                                         with_labels=True,
                                         edge_labels=lambda edge: edge.__class__.__name__)

        plt.title("World Kinematic Structure")
        plt.axis('off')  # Hide axes
        plt.show()

    def _travel_branch(self, body: Body, visitor: rustworkx.visit.DFSVisitor) -> None:
        """
        Apply a DFS Visitor to a subtree of the kinematic structure.

        :param body: Starting point of the search
        :param visitor: This visitor to apply.
        """
        rx.dfs_search(self.kinematic_structure, [body.index], visitor)

    def compile_forward_kinematics_expressions(self) -> None:
        """
        Traverse the kinematic structure and compile forward kinematics expressions for fast evaluation.
        """
        new_fks = ForwardKinematicsVisitor(self)
        self._travel_branch(self.root, new_fks)
        new_fks.compile_forward_kinematics()
        self._fk_computer = new_fks

    def _recompute_forward_kinematics(self) -> None:
        self._fk_computer.recompute()

    @copy_lru_cache()
    def compose_forward_kinematics_expression(self, root: Body, tip: Body) -> cas.TransformationMatrix:
        """
        :param root: The root body in the kinematic chain.
            It determines the starting point of the forward kinematics calculation.
        :param tip: The tip body in the kinematic chain.
            It determines the endpoint of the forward kinematics calculation.
        :return: An expression representing the computed forward kinematics of the tip body relative to the root body.
        """

        fk = cas.TransformationMatrix()
        root_chain, tip_chain = self.compute_split_chain_of_connections(root, tip)
        connection: Connection
        for connection in root_chain:
            tip_T_root = connection.origin_expression.inverse()
            fk = fk.dot(tip_T_root)
        for connection in tip_chain:
            fk = fk.dot(connection.origin_expression)
        fk.reference_frame = root
        fk.child_frame = tip
        return fk

    def compute_forward_kinematics(self, root: Body, tip: Body) -> cas.TransformationMatrix:
        """
        Compute the forward kinematics from the root body to the tip body.

        Calculate the transformation matrix representing the pose of the
        tip body relative to the root body.

        :param root: Root body for which the kinematics are computed.
        :param tip: Tip body to which the kinematics are computed.
        :return: Transformation matrix representing the relative pose of the tip body with respect to the root body.
        """
        return cas.TransformationMatrix(self.compute_forward_kinematics_np(root, tip))

    def compute_forward_kinematics_np(self, root: Body, tip: Body) -> NpMatrix4x4:
        """
        Compute the forward kinematics from the root body to the tip body, root_T_tip and return it as a 4x4 numpy ndarray.

        Calculate the transformation matrix representing the pose of the
        tip body relative to the root body, expressed as a numpy ndarray.

        :param root: Root body for which the kinematics are computed.
        :param tip: Tip body to which the kinematics are computed.
        :return: Transformation matrix representing the relative pose of the tip body with respect to the root body.
        """
        return self._fk_computer.compute_forward_kinematics_np(root, tip).copy()

    def transform(self, spatial_object: cas.SpatialType, target_frame: Body) -> cas.SpatialType:
        """
        Transform a given spatial object from its reference frame to a target frame.

        Calculate the transformation from the reference frame of the provided
        spatial object to the specified target frame. Apply the transformation
        differently depending on the type of the spatial object:

        - If the object is a Quaternion, compute its rotation matrix, transform it, and
          convert back to a Quaternion.
        - For other types, apply the transformation matrix directly.

        :param spatial_object: The spatial object to be transformed.
        :param target_frame: The target body frame to which the spatial object should
            be transformed.
        :return: The spatial object transformed to the target frame. If the input object
            is a Quaternion, the returned object is a Quaternion. Otherwise, it is the
            transformed spatial object.
        """
        target_frame_T_reference_frame = self.compute_forward_kinematics(root=target_frame,
                                                                         tip=spatial_object.reference_frame)
        if isinstance(spatial_object, cas.Quaternion):
            reference_frame_R = spatial_object.to_rotation_matrix()
            target_frame_R = target_frame_T_reference_frame @ reference_frame_R
            return target_frame_R.to_quaternion()
        else:
            return target_frame_T_reference_frame @ spatial_object

    def find_dofs_for_position_symbols(self, symbols: List[cas.Symbol]) -> List[DegreeOfFreedom]:
        result = []
        for s in symbols:
            for dof in self.degrees_of_freedom:
                if s == dof.symbols.position:
                    result.append(dof)
        return result

    def compute_inverse_kinematics(self, root: Body, tip: Body, target: cas.TransformationMatrix,
                                   dt: float = 0.05, max_iterations: int = 200,
                                   translation_velocity: float = 0.2, rotation_velocity: float = 0.2) \
            -> Dict[DegreeOfFreedom, float]:
        """
        Compute inverse kinematics using quadratic programming.

        :param root: Root body of the kinematic chain
        :param tip: Tip body of the kinematic chain
        :param target: Desired tip pose relative to the root body
        :param dt: Time step for integration
        :param max_iterations: Maximum number of iterations
        :param translation_velocity: Maximum translation velocity
        :param rotation_velocity: Maximum rotation velocity
        :return: Dictionary mapping DOF names to their computed positions
        """
        ik_solver = InverseKinematicsSolver(self)
        return ik_solver.solve(root, tip, target, dt, max_iterations, translation_velocity, rotation_velocity)

    def apply_control_commands(self, commands: np.ndarray, dt: float, derivative: Derivatives) -> None:
        """
        Updates the state of a system by applying control commands at a specified derivative level,
        followed by backward integration to update lower derivatives.

        :param commands: Control commands to be applied at the specified derivative
            level. The array length must match the number of free variables
            in the system.
        :param dt: Time step used for the integration of lower derivatives.
        :param derivative: The derivative level to which the control commands are
            applied.
        :return: None
        """
        if len(commands) != len(self.degrees_of_freedom):
            raise ValueError(
                f"Commands length {len(commands)} does not match number of free variables {len(self.degrees_of_freedom)}")

        self.state.set_derivative(derivative, commands)

        for i in range(derivative - 1, -1, -1):
            self.state.set_derivative(i, self.state.get_derivative(i) + self.state.get_derivative(i + 1) * dt)
        for connection in self.connections:
            if isinstance(connection, HasUpdateState):
                connection.update_state(dt)
        self.notify_state_change()

    def set_positions_1DOF_connection(self, new_state: Dict[Has1DOFState, float]) -> None:
        for connection, value in new_state.items():
            connection.position = value
        self.notify_state_change()



import keyword

import casadi as ca
import hypothesis.strategies as st
import numpy as np
from hypothesis import assume
from hypothesis.strategies import composite
from numpy import pi

from semantic_world.spatial_types.math import shortest_angular_distance
from semantic_world.spatial_types.spatial_types import Expression, if_else

# from giskardpy.middleware import get_middleware
# from giskardpy.utils.math import shortest_angular_distance

BIG_NUMBER = 1e100
SMALL_NUMBER = 1e-100


def vector(x):
    return st.lists(float_no_nan_no_inf(), min_size=x, max_size=x)


def angle_positive():
    return st.floats(0, 2 * np.pi)


def random_angle():
    return st.floats(-np.pi, np.pi)


def compare_axis_angle(actual_angle, actual_axis, expected_angle, expected_axis, decimal=3):
    try:
        np.testing.assert_array_almost_equal(actual_axis, expected_axis, decimal=decimal)
        np.testing.assert_almost_equal(shortest_angular_distance(actual_angle, expected_angle), 0, decimal=decimal)
    except AssertionError:
        try:
            np.testing.assert_array_almost_equal(actual_axis, -expected_axis, decimal=decimal)
            np.testing.assert_almost_equal(shortest_angular_distance(actual_angle, abs(expected_angle - 2 * pi)), 0,
                                           decimal=decimal)
        except AssertionError:
            np.testing.assert_almost_equal(shortest_angular_distance(actual_angle, 0), 0, decimal=decimal)
            np.testing.assert_almost_equal(shortest_angular_distance(0, expected_angle), 0, decimal=decimal)
            assert not np.any(np.isnan(actual_axis))
            assert not np.any(np.isnan(expected_axis))


def compare_orientations(actual_orientation: np.ndarray, desired_orientation: np.ndarray, decimal: int = 2) -> None:
    q1 = actual_orientation
    q2 = desired_orientation
    try:
        np.testing.assert_almost_equal(q1[0], q2[0], decimal=decimal)
        np.testing.assert_almost_equal(q1[1], q2[1], decimal=decimal)
        np.testing.assert_almost_equal(q1[2], q2[2], decimal=decimal)
        np.testing.assert_almost_equal(q1[3], q2[3], decimal=decimal)
    except:
        np.testing.assert_almost_equal(q1[0], -q2[0], decimal=decimal)
        np.testing.assert_almost_equal(q1[1], -q2[1], decimal=decimal)
        np.testing.assert_almost_equal(q1[2], -q2[2], decimal=decimal)
        np.testing.assert_almost_equal(q1[3], -q2[3], decimal=decimal)


@composite
def variable_name(draw):
    variable = draw(st.text('qwertyuiopasdfghjklzxcvbnm', min_size=1))
    assume(variable not in keyword.kwlist)
    return variable


@composite
def lists_of_same_length(draw, data_types=(), min_length=1, max_length=10, unique=False):
    length = draw(st.integers(min_value=min_length, max_value=max_length))
    lists = []
    for elements in data_types:
        lists.append(draw(st.lists(elements, min_size=length, max_size=length, unique=unique)))
    return lists


@composite
def rnd_joint_state(draw, joint_limits):
    return {jn: draw(st.floats(ll, ul, allow_nan=False, allow_infinity=False)) for jn, (ll, ul) in joint_limits.items()}


@composite
def rnd_joint_state2(draw, joint_limits):
    muh = draw(joint_limits)
    muh = {jn: ((ll if ll is not None else pi * 2), (ul if ul is not None else pi * 2)) for (jn, (ll, ul)) in
           muh.items()}
    return {jn: draw(st.floats(ll, ul, allow_nan=False, allow_infinity=False)) for jn, (ll, ul) in muh.items()}


def float_no_nan_no_inf(outer_limit=1e5):
    return float_no_nan_no_inf_min_max(-outer_limit, outer_limit)


def float_no_nan_no_inf_min_max(min_value=-1e5, max_value=1e5):
    return st.floats(allow_nan=False, allow_infinity=False, max_value=max_value, min_value=min_value,
                     allow_subnormal=False)


@composite
def sq_matrix(draw):
    i = draw(st.integers(min_value=1, max_value=10))
    i_sq = i ** 2
    l = draw(st.lists(float_no_nan_no_inf(outer_limit=1000), min_size=i_sq, max_size=i_sq))
    return np.array(l).reshape((i, i))


def unit_vector(length, elements=None):
    if elements is None:
        elements = float_no_nan_no_inf()
    vector = st.lists(elements, min_size=length, max_size=length).filter(
        lambda x: SMALL_NUMBER < np.linalg.norm(x) < BIG_NUMBER)

    def normalize(v):
        v = [round(x, 4) for x in v]
        l = np.linalg.norm(v)
        if l == 0:
            return np.array([0] * (length - 1) + [1])
        return np.array([x / l for x in v])

    return st.builds(normalize, vector)


def quaternion():
    return unit_vector(4, float_no_nan_no_inf(outer_limit=1))



from ..world import World
from typing_extensions import Any, Dict
from ripple_down_rules.datastructures.case import Case, create_case
from ripple_down_rules.helpers import general_rdr_classify
from . import world_views_mcrdr as views_classifier

name = 'world'
case_type = World
case_name = 'World'
classifiers_dict = dict()
classifiers_dict['views'] = views_classifier


def classify(case: World, **kwargs) -> Dict[str, Any]:
    if not isinstance(case, Case):
        case = create_case(case, max_recursion_idx=3)
    return general_rdr_classify(classifiers_dict, case, **kwargs)



from semantic_world.reasoner import WorldReasoner
from semantic_world.adapters.urdf import URDFParser
from semantic_world.views.views import Drawer


def create_kitchen_world(kitchen_path: str = '../resources/urdf/kitchen-small.urdf'):
    return URDFParser(kitchen_path).parse()


kitchen_world = create_kitchen_world()
reasoner = WorldReasoner(kitchen_world)

reasoner.fit_views([Drawer], update_existing_views=True, world_factory=create_kitchen_world)



from __future__ import annotations
from typing_extensions import List, TYPE_CHECKING

from .prefixed_name import PrefixedName

if TYPE_CHECKING:
    from .world_entity import View


class LogicalError(Exception):
    """
    An error that happens due to mistake in the logical operation or usage of the API during runtime.
    """
    ...


class UsageError(LogicalError):
    """
    An exception raised when an incorrect usage of the API is encountered.
    """
    ...


class AddingAnExistingViewError(UsageError):
    def __init__(self, view: View):
        msg = f'View {view} already exists.'
        super().__init__(msg)


class DuplicateViewError(UsageError):
    def __init__(self, views: List[View]):
        msg = f'Views {views} are duplicates, while views elements should be unique.'
        super().__init__(msg)


class ViewNotFoundError(UsageError):
    def __init__(self, name: PrefixedName):
        msg = f'View with name {name} not found'
        super().__init__(msg)



from __future__ import annotations

from dataclasses import dataclass, field
from os.path import dirname

from typing_extensions import ClassVar, List, Dict, Any, TYPE_CHECKING, Optional, Callable, Type

from ripple_down_rules import GeneralRDR, CaseQuery
from semantic_world.world_entity import View

if TYPE_CHECKING:
    from semantic_world.world import World


@dataclass
class WorldReasoner:
    """
    The world reasoner is a class that uses Ripple Down Rules to reason on world related concepts like world views.
    The reasoner can be used in two ways:
    1. Classification mode, where the reasoner infers all concepts that it has rules for at that time.
    >>> reasoner = WorldReasoner(world)
    >>> inferred_concepts = reasoner.reason()
    >>> inferred_views = inferred_concepts['views']
    2. Fitting mode, where the reasoner prompts the expert for answers given a query on a world concept. This allows
    incremental knowledge gain, improved reasoning capabilities, and an increased breadth of application with more
     usage.
     >>> reasoner = WorldReasoner(world)
     >>> reasoner.fit_attribute("views", [Handle, Drawer])
    """
    world: World
    """
    The semantic world instance on which the reasoning is performed.
    """
    rdr: ClassVar[GeneralRDR] = GeneralRDR(save_dir=dirname(__file__), model_name="world_rdr")
    """
    This is a ripple down rules reasoner that infers concepts relevant to the world like finding out the views that
    exist in the current world.
    """
    _last_world_model_version: Optional[int] = field(init=False, default=None)
    """
    The last world model version of the world used when :py:meth:`reason` 
    was last called.
    """
    result: Optional[Dict[str, Any]] = field(init=False, default=None)
    """
    The latest result of the :py:meth:`reason` call.
    """

    def infer_views(self) -> List[View]:
        """
        Infer the views of the world by calling the :py:meth:`reason` method and extracting all inferred views.

        :return: The inferred views of the world.
        """
        result = self.reason()
        if 'views' in result:
            views = result['views']
        else:
            views = []
        return views

    def reason(self) -> Dict[str, Any]:
        """
        Perform rule-based reasoning on the current world and infer all possible concepts.

        :return: The inferred concepts as a dictionary mapping concept name to all inferred values of that concept.
        """
        if self.world._model_version != self._last_world_model_version:
            self.result = self.rdr.classify(self.world)
            self._update_world_attributes()
            self._last_world_model_version = self.world._model_version
        return self.result

    def _update_world_attributes(self):
        """
        Update the world attributes from the values in the result of the latest :py:meth:`reason` call.
        """
        for attr_name, attr_value in self.result.items():
            if isinstance(getattr(self.world, attr_name), list):
                attr_value = list(attr_value)
            if attr_name == 'views':
                for view in attr_value:
                    self.world.add_view(view, exists_ok=True)
            else:
                setattr(self.world, attr_name, attr_value)

    def fit_views(self, required_views: List[Type[View]],
                  update_existing_views: bool = False,
                  world_factory: Optional[Callable] = None,
                  scenario: Optional[Callable] = None) -> None:
        """
        Fit the world RDR to the required view types.

        :param required_views: A list of view types that the RDR should be fitted to.
        :param update_existing_views: If True, existing views will be updated with new rules, else they will be skipped.
        :param world_factory: Optional callable that can be used to recreate the world object.
        :param scenario: Optional callable that represents the test method or scenario that is being executed.
        """
        self.fit_attribute("views", required_views, update_existing_rules=update_existing_views,
                           world_factory=world_factory, scenario=scenario)

    def fit_attribute(self, attribute_name: str, attribute_types: List[Type[Any]], update_existing_rules: bool = False,
                      world_factory: Optional[Callable] = None,
                      scenario: Optional[Callable] = None) -> None:
        """
        Fit the world RDR to the required attribute types.

        :param attribute_name: The attribute name that the RDR should be fitted to.
        :param attribute_types: A list of attribute types that the RDR should be fitted to.
        :param update_existing_rules: If True, existing rules of the given types will be updated with new rules,
         else they will be skipped.
        :param world_factory: Optional callable that can be used to recreate the world object.
        :param scenario: Optional callable that represents the test method or scenario that is being executed.
        """
        for attr_type in attribute_types:
            case_query = CaseQuery(self.world, attribute_name, (attr_type,), False, case_factory=world_factory,
                                   scenario=scenario)
            self.rdr.fit_case(case_query, update_existing_rules=update_existing_rules)



from dataclasses import dataclass
from typing import Optional


@dataclass
class PrefixedName:
    name: str
    prefix: Optional[str] = None

    def __hash__(self):
        return hash((self.prefix, self.name))

    def __str__(self):
        return f"{self.prefix}/{self.name}"

    def __eq__(self, other):
        return str(self) == str(other)


from typing_extensions import List, Union

from semantic_world.views.views import Cabinet, Container, Door, Drawer, Fridge, Handle
from semantic_world.connections import FixedConnection, PrismaticConnection, RevoluteConnection
from semantic_world.world import World


def conditions_90574698325129464513441443063592862114(case) -> bool:
    def has_bodies_named_handle(case: World) -> bool:
        """Get conditions on whether it's possible to conclude a value for World.views  of type Handle."""
        return any("handle" in b.name.name.lower() for b in case.bodies)

    return has_bodies_named_handle(case)


def conclusion_90574698325129464513441443063592862114(case) -> List[Handle]:
    def get_handles(case: World) -> Union[set, list, Handle]:
        """Get possible value(s) for World.views of types list/set of Handle"""
        return [Handle(b) for b in case.bodies if "handle" in b.name.name.lower()]

    return get_handles(case)


def conditions_14920098271685635920637692283091167284(case) -> bool:
    def has_handles_and_fixed_and_prismatic_connections(case: World) -> bool:
        """Get conditions on whether it's possible to conclude a value for World.views  of type Container."""
        return (any(v for v in case.views if type(v) is Handle) and
                any(c for c in case.connections if isinstance(c, PrismaticConnection)) and
                any(c for c in case.connections if isinstance(c, FixedConnection)))

    return has_handles_and_fixed_and_prismatic_connections(case)


def conclusion_14920098271685635920637692283091167284(case) -> List[Container]:
    def get_containers(case: World) -> Union[set, Container, list]:
        """Get possible value(s) for World.views of types list/set of Container"""
        prismatic_connections = [c for c in case.connections if isinstance(c, PrismaticConnection)]
        fixed_connections = [c for c in case.connections if isinstance(c, FixedConnection)]
        children_of_prismatic_connections = [c.child for c in prismatic_connections]
        handles = [v for v in case.views if type(v) is Handle]
        fixed_connections_with_handle_child = [fc for fc in fixed_connections if fc.child in [h.body for h in handles]]
        drawer_containers = set(children_of_prismatic_connections).intersection(
            set([fc.parent for fc in fixed_connections_with_handle_child]))
        return [Container(b) for b in drawer_containers]

    return get_containers(case)


def conditions_331345798360792447350644865254855982739(case) -> bool:
    def has_handles_and_containers(case: World) -> bool:
        """Get conditions on whether it's possible to conclude a value for World.views  of type Drawer."""
        return any(v for v in case.views if type(v) is Handle) and any(v for v in case.views if type(v) is Container)

    return has_handles_and_containers(case)


def conclusion_331345798360792447350644865254855982739(case) -> List[Drawer]:
    def get_drawers(case: World) -> Union[set, list, Drawer]:
        """Get possible value(s) for World.views of types list/set of Drawer"""
        handles = [v for v in case.views if type(v) is Handle]
        containers = [v for v in case.views if type(v) is Container]
        fixed_connections = [c for c in case.connections if
                             isinstance(c, FixedConnection) and c.parent in [cont.body for cont in
                                                                             containers] and c.child in [
                                 h.body for h in handles]]
        prismatic_connections = [c for c in case.connections if
                                 isinstance(c, PrismaticConnection) and c.child in [cont.body for cont in containers]]
        drawer_handle_connections = [fc for fc in fixed_connections if
                                     fc.parent in [pc.child for pc in prismatic_connections]]
        drawers = [Drawer([cont for cont in containers if dc.parent == cont.body][0],
                          [h for h in handles if dc.child == h.body][0]) for dc in drawer_handle_connections]
        return drawers

    return get_drawers(case)


def conditions_35528769484583703815352905256802298589(case) -> bool:
    def has_drawers(case: World) -> bool:
        """Get conditions on whether it's possible to conclude a value for World.views  of type Cabinet."""
        return any(v for v in case.views if type(v) is Drawer)

    return has_drawers(case)


def conclusion_35528769484583703815352905256802298589(case) -> List[Cabinet]:
    def get_cabinets(case: World) -> Union[set, Cabinet, list]:
        """Get possible value(s) for World.views of types list/set of Cabinet"""
        drawers = [v for v in case.views if type(v) is Drawer]
        prismatic_connections = [c for c in case.connections if
                                 isinstance(c, PrismaticConnection) and c.child in [drawer.container.body for drawer in
                                                                                    drawers]]
        cabinet_container_bodies = [pc.parent for pc in prismatic_connections]
        cabinets = []
        for ccb in cabinet_container_bodies:
            if ccb in [cabinet.container.body for cabinet in cabinets]:
                continue
            cc_prismatic_connections = [pc for pc in prismatic_connections if pc.parent is ccb]
            cabinet_drawer_container_bodies = [pc.child for pc in cc_prismatic_connections]
            cabinet_drawers = [d for d in drawers if d.container.body in cabinet_drawer_container_bodies]
            cabinets.append(Cabinet(Container(ccb), cabinet_drawers))

        return cabinets

    return get_cabinets(case)


def conditions_59112619694893607910753808758642808601(case) -> bool:
    def has_handles_and_revolute_connections(case: World) -> bool:
        """Get conditions on whether it's possible to conclude a value for World.views  of type Door."""
        return (any(v for v in case.views if isinstance(v, Handle)) and
                any(c for c in case.connections if isinstance(c, RevoluteConnection)))

    return has_handles_and_revolute_connections(case)


def conclusion_59112619694893607910753808758642808601(case) -> List[Door]:
    def get_doors(case: World) -> List[Door]:
        """Get possible value(s) for World.views  of type Door."""
        handles = [v for v in case.views if isinstance(v, Handle)]
        handle_bodies = [h.body for h in handles]
        connections_with_handles = [c for c in case.connections if isinstance(c, FixedConnection) and
                                    c.child in handle_bodies]

        revolute_connections = [c for c in case.connections if isinstance(c, RevoluteConnection)]
        bodies_connected_to_handles = [c.parent if c.child in handle_bodies else c.child for c in
                                       connections_with_handles]
        bodies_that_have_revolute_joints = [b for b in bodies_connected_to_handles for c in revolute_connections
                                            if b == c.child]
        body_handle_connections = [c for c in connections_with_handles if c.parent in bodies_that_have_revolute_joints]
        doors = [Door(c.parent, [h for h in handles if h.body == c.child][0]) for c in body_handle_connections]
        return doors

    return get_doors(case)


def conditions_10840634078579061471470540436169882059(case) -> bool:
    def has_doors_with_fridge_in_their_name(case: World) -> bool:
        """Get conditions on whether it's possible to conclude a value for World.views  of type Fridge."""
        return any(v for v in case.views if isinstance(v, Door) and "fridge" in v.body.name.name.lower())

    return has_doors_with_fridge_in_their_name(case)


def conclusion_10840634078579061471470540436169882059(case) -> List[Fridge]:
    def get_fridges(case: World) -> List[Fridge]:
        """Get possible value(s) for World.views  of type Fridge."""
        # Get fridge-related doors
        fridge_doors = [v for v in case.views if isinstance(v, Door) and "fridge" in v.body.name.name.lower()]
        # Precompute bodies of the fridge doors
        fridge_doors_bodies = [d.body for d in fridge_doors]
        # Filter relevant revolute connections
        fridge_door_connections = [
            c for c in case.connections
            if isinstance(c, RevoluteConnection)
               and c.child in fridge_doors_bodies
               and 'fridge' in c.parent.name.name.lower()
        ]
        return [Fridge(c.parent, fridge_doors[fridge_doors_bodies.index(c.child)]) for c in fridge_door_connections]

    return get_fridges(case)



import os
from enum import Enum

import semantic_world.degree_of_freedom
import semantic_world.geometry
import semantic_world.views.views
import semantic_world.world_entity
from ormatic.ormatic import ORMatic
from ormatic.utils import classes_of_module, recursive_subclasses
from semantic_world.connections import FixedConnection
from semantic_world.orm.model import *
from semantic_world.prefixed_name import PrefixedName
from semantic_world.world import *

# ----------------------------------------------------------------------------------------------------------------------
# This script generates the ORM classes for the semantic_world package.
# Dataclasses can be mapped automatically to the ORM model
# using the ORMatic library, they just have to be registered in the classes list.
# Classes that are self_mapped and explicitly_mapped are already mapped in the model.py file. Look there for more
# information on how to map them.
# ----------------------------------------------------------------------------------------------------------------------

# create of classes that should be mapped
classes = set(recursive_subclasses(AlternativeMapping))
classes |= set(classes_of_module(semantic_world.geometry))
classes |= set(classes_of_module(semantic_world.world))
classes |= set(classes_of_module(semantic_world.prefixed_name))
classes |= set(classes_of_module(semantic_world.world_entity))
classes |= set(classes_of_module(semantic_world.connections))
classes |= set(classes_of_module(semantic_world.views.views))
classes |= set(classes_of_module(semantic_world.degree_of_freedom))

# remove classes that should not be mapped
classes -= {ResetStateContextManager, WorldModelUpdateContextManager, HasUpdateState,
            World, ForwardKinematicsVisitor, Has1DOFState, DegreeOfFreedom}
classes -= set(recursive_subclasses(Enum))


def generate_orm():
    """
    Generate the ORM classes for the pycram package.
    """

    # Create an ORMatic object with the classes to be mapped
    ormatic = ORMatic(list(classes))

    # Generate the ORM classes
    ormatic.make_all_tables()

    path = os.path.abspath(os.path.join(os.getcwd(), '../src/semantic_world/orm/'))
    with open(os.path.join(path, 'ormatic_interface.py'), 'w') as f:
        ormatic.to_sqlalchemy_file(f)


if __name__ == '__main__':
    generate_orm()
